# -*- coding: utf-8 -*-
"""
Created on Wed Oct 13 14:24:48 2021

@author: kvidal
"""

from scipy import signal, special
import numpy as np

def time_domain_filter(data,time, freq_low, freq_high):
    """
    Modified for python from SPEDAS's function of the same name
    Purpose: band-pass filter of data, assuming constant dt between points
    Parameters:
        data: input nx3 array
        time: in seconds
        freq_low: low coutoff frequency in Hz
        freq_high: high cutoff frequency in Hz
    Returns: nx3 array of band-pass filtered data
    """
    dt = time[1]-time[0]
    nyquist = 1./(2.*dt)
    flow = freq_low/nyquist
    fhigh = freq_high/nyquist
    A = 120. # from Ergun's fa_fields_filter
    if flow == 0.0:
        f = fhigh
    else:
        f = flow
    nterms = int(5./f)
    if nterms > 5000.:
        nterms = 5000.
    out = digital_filter(flow,fhigh,A,nterms)
    new_series_x = signal.convolve(data[:,0],out,mode='same',method='direct')
    new_series_y = signal.convolve(data[:,1],out,mode='same',method='direct')
    new_series_z = signal.convolve(data[:,2],out,mode='same',method='direct')
    new_series =  np.transpose(np.vstack((new_series_x,new_series_y,new_series_z)))

    return new_series

def digital_filter(flow,fhigh,aGibbs,nterms):
    
    if fhigh < flow:
        fstop = float(1)
    else:
        fstop = float(0)
    
    # Computes Kaiser weights W(N,K) for digital filters
    # W = coef = returned array of Kaiser weights
    # N = value of N in W(N,K), ie number of terms
    # A = Size of gibbs phenomenon wiggles in -DB
    
    if aGibbs <= 21 :
        alpha = 0.
    elif (aGibbs >= 50) :
        alpha = 0.1102*(aGibbs-8.7)
    else:
        alpha = 0.5842*(aGibbs-21)**(0.4) + 0.07886*(aGibbs-21)
        
    arg = (np.arange(nterms)+1)/nterms
    coef = special.iv(0,alpha*np.sqrt(1.-arg**2))/special.iv(0,alpha)
    t = (np.arange(nterms)+1)*np.pi
    coef = coef*(np.sin(t*fhigh)-np.sin(t*flow))/t
    coef = np.concatenate((np.flip(coef), [fhigh - flow + fstop], coef))
    return coef




import os
import logging
from pyspedas.mms.mms_config import CONFIG
from pyspedas.utilities.download import download
from pytplot import get_data, cdf_to_tplot

logging.captureWarnings(True)
logging.basicConfig(format='%(asctime)s: %(message)s', datefmt='%d-%b-%y %H:%M:%S', level=logging.INFO)

def mms_part_des_photoelectrons(dist_var):
    """
    Loads and returns the FPI/DES photoelectron model based on stepper ID

    Input
    ----------
        dist_var: str
            tplot variable containing DES distribution data

    Notes
    ----------
        For more information on the model, see:
        https://agupubs.onlinelibrary.wiley.com/doi/full/10.1002/2017JA024518

    Returns
    ----------
        Dictionary containing the photoelectron model
    """
    data = get_data(dist_var)

    if data is None:
        logging.error('Problem reading DES distribution variable')
        return

    # get the metadata for the 'energy_table_name' from the global attributes
    metadata = get_data(dist_var, metadata=True)

    try:
        table_name = metadata['CDF']['GATT']['Energy_table_name']
    except KeyError:
        logging.error('Problem extracting the energy table name from the DF metadata')
        return

    stepper_id = table_name.replace('.txt', '').split('energies_des_')[-1]

    # we'll need the data rate
    data_rate = dist_var.split('_')[-1]

    # download the model file from the SDC
    pe_model = download(
        last_version=True,
        remote_path='https://lasp.colorado.edu/mms/sdc/public/data/models/fpi/',
        remote_file='mms_fpi_'+data_rate+'_l2_des-bgdist_v?.?.?_p'+stepper_id+'.cdf',
        local_path=os.path.join(CONFIG['local_data_dir'], 'mms'+os.path.sep+'sdc'+os.path.sep+'public'+os.path.sep+'data'+os.path.sep+'models'+os.path.sep+'fpi'+os.path.sep+''))

    if len(pe_model) != 1:
        logging.error('Problem downloading DES model from the SDC')
        return

    model_vars = cdf_to_tplot(pe_model[0], get_support_data=True)

    if data_rate == 'fast':
        bg_dist = get_data('mms_des_bgdist_fast')
        nphoto = get_data('mms_des_numberdensity_fast')
        return {'bg_dist': bg_dist, 'n': nphoto}
    elif data_rate == 'brst':
        bg_dist_0 = get_data('mms_des_bgdist_p0_brst')
        bg_dist_1 = get_data('mms_des_bgdist_p1_brst')
        nphoto_0 = get_data('mms_des_numberdensity_p0_brst')
        nphoto_1 = get_data('mms_des_numberdensity_p1_brst')
        return {'bgdist_p0': bg_dist_0, 
                'bgdist_p1': bg_dist_1, 
                'n_0': nphoto_0, 
                'n_1': nphoto_1}

    # shouldn't get here
    logging.error('Error: something went wrong with the photoelectron model')
    return



import os
import unittest
from pyspedas.utilities.data_exists import data_exists
import pyspedas
from pyspedas import time_double
from pyspedas.geopack import tt89
from pyspedas.geopack import tt96
from pyspedas.geopack import tt01
from pyspedas.geopack import tts04
from pyspedas.geopack.get_tsy_params import get_tsy_params
from pyspedas import tinterpol
from pytplot import join_vec

trange = ['2015-10-16', '2015-10-17']

def get_params(model):
    support_trange = [time_double(trange[0])-60*60*24, 
                      time_double(trange[1])+60*60*24]
    pyspedas.kyoto.dst(trange=support_trange)
    pyspedas.omni.data(trange=trange)
    join_vec(['BX_GSE', 'BY_GSM', 'BZ_GSM'])
    return get_tsy_params('kyoto_dst', 
                    'BX_GSE-BY_GSM-BZ_GSM_joined', 
                    'proton_density', 
                    'flow_speed', 
                    model, 
                    pressure_tvar='Pressure',
                    speed=True)

class LoadTestCases(unittest.TestCase):
    def test_igrf(self):
        mec_vars = pyspedas.mms.mec(trange=trange)
        tt89('mms1_mec_r_gsm', igrf_only=True)
        self.assertTrue(data_exists('mms1_mec_r_gsm_bt89'))

    def test_tt89(self):
        mec_vars = pyspedas.mms.mec(trange=trange)
        tt89('mms1_mec_r_gsm')
        self.assertTrue(data_exists('mms1_mec_r_gsm_bt89'))

    def test_tt96(self):
        mec_vars = pyspedas.mms.mec(trange=trange)
        params = get_params('t96')
        tinterpol('mms1_mec_r_gsm', 'proton_density')
        tt96('mms1_mec_r_gsm-itrp', parmod=params)
        self.assertTrue(data_exists('mms1_mec_r_gsm-itrp_bt96'))

    def test_tt01(self):
        mec_vars = pyspedas.mms.mec(trange=trange)
        params = get_params('t01')
        tinterpol('mms1_mec_r_gsm', 'proton_density')
        tt01('mms1_mec_r_gsm-itrp', parmod=params)
        self.assertTrue(data_exists('mms1_mec_r_gsm-itrp_bt01'))

    def test_tts04(self):
        mec_vars = pyspedas.mms.mec(trange=trange)
        params = get_params('ts04')
        tinterpol('mms1_mec_r_gsm', 'proton_density')
        tts04('mms1_mec_r_gsm-itrp', parmod=params)
        self.assertTrue(data_exists('mms1_mec_r_gsm-itrp_bts04'))

if __name__ == '__main__':
    unittest.main()



import numpy as np

def moments_3d_omega_weights(theta, phi, dtheta, dphi):
    """
    Helper function used by moments_3d

    Input:
        theta: numpy.ndarray
            theta bin values

        phi: numpy.ndarray
            phi bin values

        dtheta: numpy.ndarray
            Widths of the theta bins
            
        dphi: numpy.ndarray
            Widths of the phi bins

    Notes:
        The calculations were heisted from Davin Larson's IDL version

    Returns:
        Omega weights to be used in moments_3d

    """

    omega = np.zeros([13, theta.shape[0], theta.shape[1]])

    # Angular moment integrals
    ph2 = phi + dphi/2.0
    ph1 = phi - dphi/2.0
    th2 = theta + dtheta/2.0
    th1 = theta - dtheta/2.0

    sth1 = np.sin(th1*np.pi/180.0)
    cth1 = np.cos(th1*np.pi/180.0)
    sph1 = np.sin(ph1*np.pi/180.0)
    cph1 = np.cos(ph1*np.pi/180.0)

    sth2 = np.sin(th2*np.pi/180.0)
    cth2 = np.cos(th2*np.pi/180.0)
    sph2 = np.sin(ph2*np.pi/180.0)
    cph2 = np.cos(ph2*np.pi/180.0)

    ip = dphi*np.pi/180.0
    ict = sth2 - sth1
    icp = sph2 - sph1
    isp = -cph2 + cph1
    is2p = dphi/2.0*np.pi/180.0 - sph2*cph2/2.0 + sph1*cph1/2.0
    ic2p = dphi/2.0*np.pi/180.0 + sph2*cph2/2.0 - sph1*cph1/2.0
    ic2t = dtheta/2.0*np.pi/180.0 + sth2*cth2/2.0 - sth1*cth1/2.0
    ic3t = sth2 - sth1 - (sth2**3 - sth1**3)/3.0
    ictst = (sth2**2 - sth1**2)/2.0
    icts2t = (sth2**3 - sth1**3)/3.0
    ic2tst = (-cth2**3 + cth1**3)/3.0
    icpsp = (sph2**2 - sph1**2)/2.0

    omega[0, :, :] = ict*ip
    omega[1, :, :] = ic2t*icp
    omega[2, :, :] = ic2t*isp
    omega[3, :, :] = ictst*ip
    omega[4, :, :] = ic3t*ic2p
    omega[5, :, :] = ic3t*is2p
    omega[6, :, :] = icts2t*ip
    omega[7, :, :] = ic3t*icpsp
    omega[8, :, :] = ic2tst*icp
    omega[9, :, :] = ic2tst*isp
    omega[10, :, :] = omega[1, :, :]
    omega[11, :, :] = omega[2, :, :]
    omega[12, :, :] = omega[3, :, :]

    return omega



import pytplot
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import LinearSegmentedColormap
from .slice2d_getinfo import slice2d_getinfo


def plot(the_slice,
         xrange=None,
         yrange=None,
         zrange=None,
         colormap='spedas',
         plotsize=10,
         save_png=None,
         save_svg=None,
         save_pdf=None,
         save_eps=None,
         display=True):
    """
    Creates plots of 2D particle slices
    """
    spec_options = {}

    if zrange is None:
        zrange = the_slice.get('zrange')
        # padding, # in thm_ui_slice2d should match this
        zrange = [zrange[0]*0.999, zrange[1]*1.001]

    if zrange is None:
        spec_options['norm'] = mpl.colors.LogNorm()
    else:
        spec_options['norm'] = mpl.colors.LogNorm(vmin=zrange[0], vmax=zrange[1])

    if colormap == 'spedas':
        _colors = pytplot.spedas_colorbar
        spd_map = [(np.array([r, g, b])).astype(np.float64) / 256 for r, g, b in zip(_colors.r, _colors.g, _colors.b)]
        cmap = LinearSegmentedColormap.from_list('spedas', spd_map)
    else:
        cmap = colormap

    spec_options['cmap'] = cmap

    fig, axes = plt.subplots()
    fig.set_size_inches(plotsize, plotsize)

    if xrange is not None:
        axes.set_xlim(xrange)

    if yrange is not None:
        axes.set_ylim(yrange)

    info = slice2d_getinfo(the_slice)

    axes.set_title(info['title'])
    axes.set_ylabel(info['ytitle'])
    axes.set_xlabel(info['xtitle'])

    fig.subplots_adjust(left=0.14, right=0.86, top=0.86, bottom=0.14)

    box = axes.get_position()
    pad, width = 0.02, 0.01
    cax = fig.add_axes([box.xmax + pad, box.ymin, width, box.height])

    im = axes.pcolormesh(the_slice['xgrid'], the_slice['ygrid'], the_slice['data'].T, **spec_options)

    colorbar = fig.colorbar(im, cax=cax)

    colorbar.set_label(info['ztitle'])

    # draw lines at the origin
    axes.axvline(x=0, linestyle=(0, (5, 10)), color='black')
    axes.axhline(y=0, linestyle=(0, (5, 10)), color='black')

    if save_png is not None and save_png != '':
        plt.savefig(save_png + '.png')

    if save_eps is not None and save_eps != '':
        plt.savefig(save_eps + '.eps')

    if save_svg is not None and save_svg != '':
        plt.savefig(save_svg + '.svg')

    if save_pdf is not None and save_pdf != '':
        plt.savefig(save_pdf + '.pdf')

    if display:
        plt.show()




import numpy as np
from scipy.ndimage.interpolation import shift

def mms_pgs_clean_data(data_in):
    """
    Sanitize MMS FPI/HPCA data structures for use with
    mms_part_products; reforms energy by theta by phi to energy by angle
    and calculates delta-energy for each bin
    """

    output = {'charge': data_in['charge'], 'mass': data_in['mass'],
              'data': np.reshape(data_in['data'], [data_in['data'].shape[0], data_in['data'].shape[1]*data_in['data'].shape[2]], order='F'),
              'bins': np.reshape(data_in['bins'], [data_in['data'].shape[0], data_in['data'].shape[1]*data_in['data'].shape[2]], order='F'),
              'theta': np.reshape(data_in['theta'], [data_in['data'].shape[0], data_in['data'].shape[1]*data_in['data'].shape[2]], order='F'),
              'energy': np.reshape(data_in['energy'], [data_in['data'].shape[0], data_in['data'].shape[1]*data_in['data'].shape[2]], order='F'),
              'phi': np.reshape(data_in['phi'], [data_in['data'].shape[0], data_in['data'].shape[1]*data_in['data'].shape[2]], order='F'),
              'dtheta': np.reshape(data_in['dtheta'], [data_in['data'].shape[0], data_in['data'].shape[1]*data_in['data'].shape[2]], order='F'),
              'dphi': np.reshape(data_in['dphi'], [data_in['data'].shape[0], data_in['data'].shape[1]*data_in['data'].shape[2]], order='F'),
              'denergy': np.reshape(data_in['denergy'], [data_in['data'].shape[0], data_in['data'].shape[1]*data_in['data'].shape[2]], order='F')}

    de = output['energy'] - shift(output['energy'], [1, 0])
    output['denergy'] = shift((de+shift(de, [1, 0]))/2.0, -1)
    # just have to make a guess at the edges(bottom edge)
    output['denergy'][0, :] = de[1, :]
    # just have to make a guess at the edges(top edge)
    output['denergy'][-1, :] = de[-1, :]

    return output


import logging
import numpy as np

from pytplot import get_data

from pyspedas.utilities.interpol import interpol
from pyspedas.particles.spd_part_products.spd_pgs_make_tplot import spd_pgs_make_tplot
from pyspedas.particles.spd_part_products.spd_pgs_limit_range import spd_pgs_limit_range
from pyspedas.particles.spd_part_products.spd_pgs_progress_update import spd_pgs_progress_update
from pyspedas.particles.spd_part_products.spd_pgs_do_fac import spd_pgs_do_fac
from pyspedas.particles.spd_part_products.spd_pgs_regrid import spd_pgs_regrid
from pyspedas.particles.moments.spd_pgs_moments import spd_pgs_moments
from pyspedas.particles.moments.spd_pgs_moments_tplot import spd_pgs_moments_tplot

from pyspedas.mms.fpi.mms_get_fpi_dist import mms_get_fpi_dist
from pyspedas.mms.hpca.mms_get_hpca_dist import mms_get_hpca_dist
from pyspedas.mms.particles.mms_convert_flux_units import mms_convert_flux_units
from pyspedas.mms.particles.mms_pgs_clean_data import mms_pgs_clean_data
from pyspedas.mms.particles.mms_pgs_clean_support import mms_pgs_clean_support
from pyspedas.mms.particles.mms_pgs_make_fac import mms_pgs_make_fac
from pyspedas.mms.particles.mms_pgs_split_hpca import mms_pgs_split_hpca
from pyspedas.mms.particles.mms_pgs_make_e_spec import mms_pgs_make_e_spec
from pyspedas.mms.particles.mms_pgs_make_phi_spec import mms_pgs_make_phi_spec
from pyspedas.mms.particles.mms_pgs_make_theta_spec import mms_pgs_make_theta_spec
from pyspedas.mms.particles.mms_part_des_photoelectrons import mms_part_des_photoelectrons

logging.captureWarnings(True)
logging.basicConfig(format='%(asctime)s: %(message)s', datefmt='%d-%b-%y %H:%M:%S', level=logging.INFO)


def mms_part_products(in_tvarname, units='eflux', species='e', data_rate='fast', instrument='fpi', probe='1',
    output=['energy', 'theta', 'phi'], energy=None, phi=None, theta=None, pitch=None, gyro=None, mag_name=None,
    pos_name=None, fac_type='mphigeo', sc_pot_name=None, correct_photoelectrons=False, zero_negative_values=False,
    internal_photoelectron_corrections=False, disable_photoelectron_corrections=False, no_regrid=False,
    regrid=[32, 16], vel_name=None):
    """
    Generate spectra and moments from 3D MMS particle data; note: this routine isn't
    meant to be called directly - see the wrapper mms_part_getspec instead.

    Input
    ----------
        in_tvarname: str
            Name of the tplot variable containing MMS 3D particle distribution data

    Parameters
    ----------
        units: str
            Specify units of output variables; must be 'eflux' to calculate moments

            valid options:
            'flux'   -  # / (cm^2 * s * sr * eV)
            'eflux'  -  eV / (cm^2 * s * sr * eV)  <default>
            'df_cm'  -  s^3 / cm^6
            'df_km'  -  s^3 / km^6

        species: str
            Specify the species of the input tplot variable

        data_rate: str
            Data rate of the input data

        instrument: str
            Instrument (FPI or HPCA)

        probe: int or str
            Spacecraft probe #

        output: str or list of str
            Output variables; options: 
                'energy': energy spectrograms
                'theta': theta spectrograms
                'phi': phi spectrograms
                'pa': pitch-angle spectrograms
                'gyro': gyro-phase spectrograms
                'moments': plasma moments

        energy: list of float
            Energy range [min, max], in eV

        phi: list of float
            Phi range [min, max], in degrees

        theta: list of float
            Theta range [min, max], in degrees

        pitch: list of float
            Pitch-angle range [min, max], in degrees

        gyro: list of float
            Gyro-angle range [min, max], in degrees

        mag_name: str
            Tplot variable containing magnetic field data for
            moments and FAC transformations

        pos_name: str
            Tplot variable containing spacecraft position for
            FAC transformations

        sc_pot_name: str
            Tplot variable containing spacecraft potential data
            for moments corrections

        fac_type: str
            Field aligned coordinate system variant; default: 'mphigeo'
            options: 'phigeo', 'mphigeo', 'xgse'

        correct_photoelectrons: bool
            Flag to correct FPI data for photoelectrons 
            (defaults to True for FPI electron data - disable with the parameter below)

        disable_photoelectron_corrections: bool
            Flag to disable FPI photoelectron corrections

        internal_photoelectron_corrections: bool
            Apply internal photoelectron corrections

        zero_negative_values: bool
            Turn negative values to 0 after doing the photoelectron corrections (DES)
            
    Returns
    ----------
        Creates tplot variables containing spectrograms and moments

    """

    units = units.lower()
    if isinstance(probe, int):
        probe = str(probe)

    data_in = get_data(in_tvarname)

    if data_in is None:
        logging.error('Error, could not find the variable: ' + in_tvarname)
        return None

    if isinstance(output, str):
        output = output.split(' ')

    if instrument == 'fpi':
        dist_in = mms_get_fpi_dist(in_tvarname, species=species, probe=probe, data_rate=data_rate)
    elif instrument == 'hpca':
        dist_in = mms_get_hpca_dist(in_tvarname, species=species, probe=probe, data_rate=data_rate)
    else:
        logging.error('Error, unknown instrument: ' + instrument + '; valid options: fpi, hpca')
        return

    if instrument == 'hpca':
        data_times = mms_get_hpca_dist(in_tvarname, species=species, probe=probe, data_rate=data_rate, times=True)
    else:
        data_times = data_in.times

    # ntimes = len(data_times)
    ntimes = len(dist_in)

    # create rotation matrix to field aligned coordinates if needed
    fac_outputs = ['pa', 'gyro', 'fac_energy', 'fac_moments']
    fac_requested = len(set(output).intersection(fac_outputs)) > 0
    if fac_requested:
        fac_matrix = mms_pgs_make_fac(data_times, mag_name, pos_name, fac_type=fac_type)

        if fac_matrix is None:
            # problem creating the FAC matrices
            fac_requested = False

    out_energy = np.zeros((ntimes, dist_in[0]['n_energy']))
    out_energy_y = np.zeros((ntimes, dist_in[0]['n_energy']))
    out_theta = np.zeros((ntimes, dist_in[0]['n_theta']))
    out_phi = np.zeros((ntimes, dist_in[0]['n_phi']))
    out_theta_y = np.zeros((ntimes, dist_in[0]['n_theta']))
    out_phi_y = np.zeros((ntimes, dist_in[0]['n_phi']))
    if fac_requested:
        out_pad = np.zeros((ntimes, dist_in[0]['n_theta']))
        out_pad_y = np.zeros((ntimes, dist_in[0]['n_theta']))
        out_gyro = np.zeros((ntimes, dist_in[0]['n_phi']))
        out_gyro_y = np.zeros((ntimes, dist_in[0]['n_phi']))

    # moments
    if 'moments' in output:
        out_density = np.zeros(ntimes)
        out_avgtemp = np.zeros(ntimes)
        out_vthermal = np.zeros(ntimes)
        out_flux = np.zeros([ntimes, 3])
        out_velocity = np.zeros([ntimes, 3])
        out_mftens = np.zeros([ntimes, 6])
        out_ptens = np.zeros([ntimes, 6])
        #out_ttens = np.zeros([dist_in['n_times'], 3, 3])

    out_vars = []
    last_update_time = None

    if 'moments' in output or correct_photoelectrons or internal_photoelectron_corrections:
        support_data = mms_pgs_clean_support(data_times, mag_name=mag_name, vel_name=vel_name, sc_pot_name=sc_pot_name)
        mag_data = support_data[0]
        scpot_data = support_data[2]

    if disable_photoelectron_corrections:
        correct_photoelectrons = False
        internal_photoelectron_corrections = False

    # grab the DES photoelectron model if needed
    if (instrument != 'fpi' or species != 'e') and (correct_photoelectrons or internal_photoelectron_corrections):
        logging.error('Photoelectron corrections only valid for DES; no corrections will be applied.')
        correct_photoelectrons = False
        internal_photoelectron_corrections = False

    if correct_photoelectrons or internal_photoelectron_corrections:
        fpi_photoelectrons = mms_part_des_photoelectrons(in_tvarname)

        if fpi_photoelectrons is None:
            logging.error('Photoelectron model missing for this date; re-run without photoelectron corrections')
            return

        # will need stepper parities for burst mode data
        if data_rate == 'brst':
            parity = get_data('mms'+probe+'_des_steptable_parity_brst')

        startdelphi = get_data('mms'+probe+'_des_startdelphi_count_'+data_rate)

    for i in range(0, ntimes):
        last_update_time = spd_pgs_progress_update(last_update_time=last_update_time, current_sample=i, total_samples=ntimes, type_string=in_tvarname)

        if instrument == 'fpi':
            dists = mms_get_fpi_dist(in_tvarname, index=i, species=species, probe=probe, data_rate=data_rate)
        elif instrument == 'hpca':
            dists = mms_get_hpca_dist(in_tvarname, index=i, species=species, probe=probe, data_rate=data_rate)

        if isinstance(dists, list):
            dist_in = dists[0]
        else:
            dist_in = dists

        # apply the DES photoelectron corrections
        if correct_photoelectrons or internal_photoelectron_corrections:
            # From Dan Gershman's release notes on the FPI photoelectron model:
            # Find the index I in the startdelphi_counts_brst or startdelphi_counts_fast array
            # [360 possibilities] whose corresponding value is closest to th = e measured
            # startdelphi_count_brst or startdelphi_count_fast for the skymap of interest. The
            # closest index can be approximated by I = floor(startdelphi_count_brst/16) or I =
            # floor(startdelphi_count_fast/16)
            startdelphi_I = int(np.floor(startdelphi.y[i]/16.0))

            if data_rate == 'brst':
                parity_num = str(int(np.fix(parity.y[i])))

                bg_dist = fpi_photoelectrons['bgdist_p'+parity_num]
                n_value = fpi_photoelectrons['n_'+parity_num]

                fphoto = bg_dist.y[startdelphi_I, :, :, :]

                # need to interpolate using SC potential data to get Nphoto value
                nphoto_scpot_dependent = n_value.y[startdelphi_I, :]
                nphoto = interpol(nphoto_scpot_dependent, n_value.v, scpot_data[i])
            else:
                fphoto = fpi_photoelectrons['bg_dist'].y[startdelphi_I, :, :, :]

                # need to interpolate using SC potential data to get Nphoto value
                nphoto_scpot_dependent = fpi_photoelectrons['n'].y[startdelphi_I, :]
                nphoto = interpol(nphoto_scpot_dependent, fpi_photoelectrons['n'].v, scpot_data[i])

            # now, the corrected distribution function is simply f_corrected = f-fphoto*nphoto
            # note: transpose is to shuffle fphoto*nphoto to energy-azimuth-elevation, to match dist.data
            correction = fphoto*nphoto
            corrected_df = dist_in['data']-correction.transpose([2, 0, 1])

            if zero_negative_values:
                corrected_df[corrected_df < 0] = 0.0

            dist_in['data'] = corrected_df

        data = mms_convert_flux_units(dist_in, units=units)

        clean_data = mms_pgs_clean_data(data)

        # split hpca angle bins to be equal width in phi/theta
        # this is needed when skipping the regrid step
        if instrument == 'hpca':
            clean_data = mms_pgs_split_hpca(clean_data)

        # Apply phi, theta, & energy limits
        if energy is not None or theta is not None or phi is not None:
            clean_data = spd_pgs_limit_range(clean_data, energy=energy, theta=theta, phi=phi)

        # Build energy spectrogram
        if 'energy' in output:
            out_energy_y[i, :], out_energy[i, :] = mms_pgs_make_e_spec(clean_data)

        # Build theta spectrogram
        if 'theta' in output:
            out_theta_y[i, :], out_theta[i, :] = mms_pgs_make_theta_spec(clean_data, resolution=dist_in['n_theta'])

        # Build phi spectrogram
        if 'phi' in output:
            out_phi_y[i, :], out_phi[i, :] = mms_pgs_make_phi_spec(clean_data, resolution=dist_in['n_phi'])

        # Calculate the moments
        if 'moments' in output:
            if scpot_data is not None:
                scpot_val = scpot_data[i]
            else:
                scpot_val = 0.0

            moments = spd_pgs_moments(clean_data, sc_pot=scpot_val)
            out_density[i] = moments['density']
            out_avgtemp[i] = moments['avgtemp']
            out_vthermal[i] = moments['vthermal']
            out_flux[i, :] = moments['flux']
            out_velocity[i, :] = moments['velocity']
            out_mftens[i, :] = moments['mftens']
            out_ptens[i, :] = moments['ptens']

        # Perform transformation to FAC, regrid data, and apply limits in new coords
        if fac_requested:
            fac_data = spd_pgs_do_fac(clean_data, fac_matrix[i, :, :])

            if no_regrid == False:
                fac_data = spd_pgs_regrid(fac_data, regrid)

            fac_data['theta'] = 90.0-fac_data['theta']
            fac_data = spd_pgs_limit_range(fac_data, theta=pitch, phi=gyro)

        if 'pa' in output:
            out_pad_y[i, :], out_pad[i, :] = mms_pgs_make_theta_spec(fac_data, colatitude=True, resolution=dist_in['n_theta'])

        if 'gyro' in output:
            out_gyro_y[i, :], out_gyro[i, :] = mms_pgs_make_phi_spec(fac_data, resolution=dist_in['n_phi'])


    if 'moments' in output:
        # put all of the moments arrays into a hash table prior to passing to the tplot routine
        moments = {'density': out_density, 
              'flux': out_flux, 
              'mftens': out_mftens, 
              'velocity': out_velocity, 
              'ptens': out_ptens,
              'vthermal': out_vthermal,
              'avgtemp': out_avgtemp}
        moments_vars = spd_pgs_moments_tplot(moments, x=data_times, prefix=in_tvarname)
        out_vars.extend(moments_vars)

    if 'energy' in output:
        spd_pgs_make_tplot(in_tvarname+'_energy', x=data_times, y=out_energy_y, z=out_energy, units=units, ylog=True, ytitle=dist_in['data_name'], ysubtitle='energy (eV)')
        out_vars.append(in_tvarname+'_energy')

    if 'theta' in output:
        spd_pgs_make_tplot(in_tvarname+'_theta', x=data_times, y=out_theta_y, z=out_theta, units=units, ytitle=dist_in['data_name'], ysubtitle='theta (deg)')
        out_vars.append(in_tvarname+'_theta')

    if 'phi' in output:
        spd_pgs_make_tplot(in_tvarname+'_phi', x=data_times, y=out_phi_y, z=out_phi, units=units, ytitle=dist_in['data_name'], ysubtitle='phi (deg)')
        out_vars.append(in_tvarname+'_phi')

    if 'pa' in output:
        spd_pgs_make_tplot(in_tvarname+'_pa', x=data_times, y=out_pad_y, z=out_pad, units=units, ytitle=dist_in['data_name'], ysubtitle='PA (deg)')
        out_vars.append(in_tvarname+'_pa')

    if 'gyro' in output:
        spd_pgs_make_tplot(in_tvarname+'_gyro', x=data_times, y=out_gyro_y, z=out_gyro, units=units, ytitle=dist_in['data_name'], ysubtitle='gyro (deg)')
        out_vars.append(in_tvarname+'_gyro')

    return out_vars


import numpy as np
from pyspedas import tinterpol
from pyspedas.analysis.tcrossp import tcrossp
from pyspedas.analysis.tnormalize import tnormalize
from pyspedas.cotrans.cotrans import cotrans
from pyspedas.utilities.time_string import time_string
from pytplot import get_data, get_timespan, options, store_data, tplot_names

from ...orb.orb import orb
from .cart_trans_matrix_make import cart_trans_matrix_make
from .erg_interpolate_att import erg_interpolate_att


def dsi2j2000(name_in=None,
              name_out=None,
              no_orb=False,
              J20002DSI=False,
              noload=False):
    """
    This function transform a time series data between the DSI and J2000 coordinate systems

    Parameters:

        name_in : str
            input tplot variable to be transformed

        name_out : str
            Name of the tplot variable in which the transformed data is stored

        J20002DSI : bool
            Set to transform data from J2000 to DSI. If not set, it transforms data from DSI to J2000.

    Returns:
        None

    """
    if (name_in is None) or (name_in not in tplot_names(quiet=True)):
        print('Input of Tplot name is undifiend')
        return

    if name_out is None:
        print('Tplot name for output is undifiend')
        name_out = 'result_of_dsi2j2000'

    # prepare for transformed Tplot Variable
    reload = not noload
    dl_in = get_data(name_in, metadata=True)
    get_data_array = get_data(name_in)
    time_array = get_data_array[0]
    time_length = time_array.shape[0]
    dat = get_data_array[1]

    # Get the SGI axis by interpolating the attitude data
    dsiz_j2000 = erg_interpolate_att(name_in, noload=noload)['sgiz_j2000']

    # Sun direction in J2000
    sundir = np.array([[1., 0., 0.]]*time_length)

    if no_orb:
        store_data('sundir_gse', data={'x': time_array, 'y': sundir})

    else:  # Calculate the sun directions from the instantaneous satellite locations
        if reload:
            tr = get_timespan(name_in)
            orb(trange=time_string([tr[0] - 60., tr[1] + 60.]))
            tinterpol('erg_orb_l2_pos_gse', time_array)
            scpos = get_data('erg_orb_l2_pos_gse-itrp')[1]
            sunpos = np.array([[1.496e+08, 0., 0.]]*time_length)
            sundir = sunpos - scpos
            store_data('sundir_gse', data={'x': time_array, 'y': sundir})
            tnormalize('sundir_gse', newname='sundir_gse')

    # Derive DSI-X and DSI-Y axis vectors in J2000.
    # The elementary vectors below are the definition of DSI. The detailed relationship
    # between the spin phase, sun pulse timing, sun direction, and the actual subsolar point
    # on the spining s/c body should be incorporated into the calculation below.
    if reload:
        cotrans(name_in='sundir_gse', name_out='sundir_j2000',
                coord_in='gse', coord_out='j2000')
    sun_j2000 = get_data('sundir_j2000')
    dsiy = tcrossp(dsiz_j2000['y'], sun_j2000[1], return_data=True)
    dsix = tcrossp(dsiy, dsiz_j2000['y'], return_data=True)
    dsix_j2000 = {'x': time_array, 'y': dsix}
    dsiy_j2000 = {'x': time_array, 'y': dsiy}

    if not J20002DSI:
        print('DSI --> J2000')
        mat = cart_trans_matrix_make(
            dsix_j2000['y'], dsiy_j2000['y'], dsiz_j2000['y'])
        j2000x_in_dsi = np.dot(mat, np.array([1., 0., 0.]))
        j2000y_in_dsi = np.dot(mat, np.array([0., 1., 0.]))
        j2000z_in_dsi = np.dot(mat, np.array([0., 0., 1.]))
        mat = cart_trans_matrix_make(
            j2000x_in_dsi, j2000y_in_dsi, j2000z_in_dsi)
        dat_new = np.einsum("ijk,ik->ij", mat, dat)
    else:
        print('J2000 --> DSI')
        mat = cart_trans_matrix_make(
            dsix_j2000['y'], dsiy_j2000['y'], dsiz_j2000['y'])
        dat_new = np.einsum("ijk,ik->ij", mat, dat)

    store_data(name_out, data={'x': time_array, 'y': dat_new}, attr_dict=dl_in)
    options(name_out, 'ytitle', '\n'.join(name_out.split('_')))



import warnings
from time import sleep
from pyspedas import time_double
from pytplot import get_data, store_data, options
import numpy as np

try:
    from hapiclient import hapi as load_hapi
except ImportError:
    print('hapiclient not found; install with: "pip install hapiclient"')

def hapi(trange=None, server=None, dataset=None, parameters='', suffix='',
         prefix='', catalog=False):
    """
    Loads data from a HAPI server into pytplot variables

    Parameters
    -----------
        trange: list of str or list of float
            Time range to load the data for

        server: str
            HAPI server to load the data from

        dataset: str
            HAPI dataset to load

        parameters: str or list of str
            Parameters in the dataset to load; default
            is to load them all

        prefix: str
            Prefix to append to the tplot variables

        suffix: str
            Suffix to append to the tplot variables

        catalog: bool
            If True, returns the server's catalog of datasets

    Returns
    -------
        List of tplot variables created.
    """

    if server is None:
        print('Error, no server specified; example servers include:')
        print('- https://cdaweb.gsfc.nasa.gov/hapi')
        print('- https://pds-ppi.igpp.ucla.edu/hapi')
        print('- http://planet.physics.uiowa.edu/das/das2Server/hapi')
        print('- https://iswa.gsfc.nasa.gov/IswaSystemWebApp/hapi')
        print('- http://lasp.colorado.edu/lisird/hapi')
        return

    if catalog:
        catalog = load_hapi(server)
        items = []
        if 'catalog' in catalog.keys():
            items = catalog['catalog']
        print('Available datasets: ')
        for item in items:
            if 'title' in item.keys():
                print(item['id'] + ': ' + item['title'])
            else:
                print(item['id'])
        return

    if dataset is None:
        print('Error, no dataset specified; please see the catalog for a list of available data sets.')
        return

    if trange is None:
        print('Error, no trange specified')
        return

    if isinstance(parameters, list):
        parameters = ','.join(parameters)

    opts = {'logging': False}

    with warnings.catch_warnings():
        warnings.simplefilter('ignore', category=ResourceWarning)
        warnings.filterwarnings('ignore', message='Unverified HTTPS request')
        data, hapi_metadata = load_hapi(server, dataset, parameters, trange[0], trange[1], **opts)

    out_vars = []

    # loop through the parameters in this dataset
    params = hapi_metadata['parameters']

    timestamps = [datapoint[0] for datapoint in data]
    unixtimes = [time_double(timestamp.decode('utf-8')) for timestamp in timestamps]

    for param_idx, param in enumerate(params[1:]):
        spec = False
        param_name = param.get('name')
        param_type = param.get('type')
        data_size = param.get('size')

        if param_type is None:
            param_type = 'double'

        if data_size is None:
            single_line = True

        try:
            if param_type == 'double':
                single_line = isinstance(data[0][param_idx+1], np.float64)
            elif param_type == 'integer':
                single_line = isinstance(data[0][param_idx+1], np.int32)
        except IndexError:
            continue

        if single_line:
            data_out = np.zeros((len(data)))
        else:
            try:
                data_out = np.zeros((len(data), len(data[0][param_idx+1])))
            except TypeError:
                continue

        for idx, datapoint in enumerate(data):
            if single_line:
                data_out[idx] = datapoint[param_idx+1]
            else:
                data_out[idx, :] = datapoint[param_idx+1]

        data_out = data_out.squeeze()

        # check for fill values
        fill_value = param.get('fill')
        if fill_value is not None:
            if param_type == 'double':
                fill_value = float(fill_value)
                data_out[data_out == fill_value] = np.nan
            elif param_type == 'integer':
                # NaN is only floating point, so we replace integer fill
                # values with 0 instead of NaN
                fill_value = int(fill_value)
                data_out[data_out == fill_value] = 0

        bins = param.get('bins')

        if bins is not None:
            centers = bins[0].get('centers')

            if centers is not None:
                spec = True

        data_table = {'x': unixtimes, 'y': data_out}

        if spec:
            data_table['v'] = centers

        saved = store_data(prefix + param_name + suffix, data=data_table)
        metadata = get_data(prefix + param_name + suffix, metadata=True)
        metadata['HAPI'] = param

        if spec:
            options(prefix + param_name + suffix, 'spec', True)

        param_units = param.get('units')
        if param_units is not None:
            options(prefix + param_name + suffix, 'ysubtitle', '[' + str(param_units) + ']')

        param_desc = param.get('description')
        if param_desc is not None:
            options(prefix + param_name + suffix, 'ytitle', param_desc)

        if saved:
            out_vars.append(prefix + param_name + suffix)

    return out_vars



import logging
import warnings
import numpy as np
from pytplot import get_data, store_data, options

# use nanmean from bottleneck if it's installed, otherwise use the numpy one
# bottleneck nanmean is ~2.5x faster
try:
    import bottleneck as bn
    nanmean = bn.nanmean
except ImportError:
    nanmean = np.nanmean

logging.captureWarnings(True)
logging.basicConfig(format='%(asctime)s: %(message)s', datefmt='%d-%b-%y %H:%M:%S', level=logging.INFO)

def mms_feeps_omni(eyes, probe='1', datatype='electron', data_units='intensity', data_rate='srvy', level='l2', suffix=''):
    """
    This function will calculate the omni-directional FEEPS spectrograms, and is automatically called from mms_load_feeps
    
    Parameters:
        eyes: dict
            Hash table containing the active sensor eyes

        probe: str
            probe #, e.g., '4' for MMS4

        datatype: str
            'electron' or 'ion'

        data_units: str
            'intensity'

        data_rate: str
            instrument data rate, e.g., 'srvy' or 'brst'

        level: str
            data level

        suffix: str
            suffix of the loaded data

    Returns:
        List of tplot variables created.
    """

    out_vars = []
    units_label = ''
    if data_units == 'intensity':
        units_label = '1/(cm^2-sr-s-keV)'
    elif data_units == 'counts':
        units_label = '[counts/s]'

    prefix = 'mms'+probe+'_epd_feeps_'
    if datatype == 'electron':
        energies = np.array([33.2, 51.90, 70.6, 89.4, 107.1, 125.2, 146.5, 171.3,
                    200.2, 234.0, 273.4, 319.4, 373.2, 436.0, 509.2])
    else:
        energies = np.array([57.9, 76.8, 95.4, 114.1, 133.0, 153.7, 177.6,
                    205.1, 236.7, 273.2, 315.4, 363.8, 419.7, 484.2,  558.6])

    # set unique energy bins per spacecraft; from DLT on 31 Jan 2017
    eEcorr = [14.0, -1.0, -3.0, -3.0]
    iEcorr = [0.0, 0.0, 0.0, 0.0]
    eGfact = [1.0, 1.0, 1.0, 1.0]
    iGfact = [0.84, 1.0, 1.0, 1.0]

    if probe == '1' and datatype == 'electron':
        energies = energies + eEcorr[0]
    if probe == '2' and datatype == 'electron':
        energies = energies + eEcorr[1]
    if probe == '3' and datatype == 'electron':
        energies = energies + eEcorr[2]
    if probe == '4' and datatype == 'electron':
        energies = energies + eEcorr[3]

    if probe == '1' and datatype == 'ion':
        energies = energies + iEcorr[0]
    if probe == '2' and datatype == 'ion':
        energies = energies + iEcorr[1]
    if probe == '3' and datatype == 'ion':
        energies = energies + iEcorr[2]
    if probe == '4' and datatype == 'ion':
        energies = energies + iEcorr[3]

    # percent error around energy bin center to accept data for averaging; 
    # anything outside of energies[i] +/- en_chk*energies[i] will be changed 
    # to NAN and not averaged   
    en_chk = 0.10

    top_sensors = eyes['top']
    bot_sensors = eyes['bottom']

    tmpdata = get_data(prefix+data_rate+'_'+level+'_'+datatype+'_top_'+data_units+'_sensorid_'+str(top_sensors[0])+'_clean_sun_removed'+suffix)

    if tmpdata is not None:
        if level != 'sitl':
            dalleyes = np.empty((len(tmpdata[0]), len(tmpdata[2]), len(top_sensors)+len(bot_sensors)))
            dalleyes[:] = np.nan

            for idx, sensor in enumerate(top_sensors):
                var_name = prefix+data_rate+'_'+level+'_'+datatype+'_top_'+data_units+'_sensorid_'+str(sensor)+'_clean_sun_removed'+suffix
                data = get_data(var_name)
                dalleyes[:, :, idx] = data[1]
                try:
                    iE = np.where(np.abs(energies-data[2]) > en_chk*energies)
                    if iE[0].size != 0:
                        dalleyes[:, iE[0], idx] = np.nan
                except Warning:
                    logging.warning('NaN in energy table encountered; sensor T' + str(sensor))
            for idx, sensor in enumerate(bot_sensors):
                var_name = prefix+data_rate+'_'+level+'_'+datatype+'_bottom_'+data_units+'_sensorid_'+str(sensor)+'_clean_sun_removed'+suffix
                data = get_data(var_name)
                dalleyes[:, :, idx+len(top_sensors)] = data[1]
                try:
                    iE = np.where(np.abs(energies-data[2]) > en_chk*energies)
                    if iE[0].size != 0:
                        dalleyes[:, iE[0], idx+len(top_sensors)] = np.nan
                except Warning:
                    logging.warning('NaN in energy table encountered; sensor B' + str(sensor))
        else: # sitl data
            dalleyes = np.empty((len(tmpdata[0]), len(tmpdata[2]), len(top_sensors)))
            dalleyes[:] = np.nan

            for idx, sensor in enumerate(top_sensors):
                var_name = prefix+data_rate+'_'+level+'_'+datatype+'_top_'+data_units+'_sensorid_'+str(sensor)+'_clean_sun_removed'+suffix
                data = get_data(var_name)
                dalleyes[:, :, idx] = data[1]
                iE = np.where(np.abs(energies-data[2]) > en_chk*energies)
                if iE[0].size != 0:
                    dalleyes[:, iE[0], idx] = np.nan

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            flux_omni = nanmean(dalleyes, axis=2)

        if probe == '1' and datatype == 'electron':
            flux_omni = flux_omni*eGfact[0]
        if probe == '2' and datatype == 'electron':
            flux_omni = flux_omni*eGfact[1]
        if probe == '3' and datatype == 'electron':
            flux_omni = flux_omni*eGfact[2]
        if probe == '4' and datatype == 'electron':
            flux_omni = flux_omni*eGfact[3]

        if probe == '1' and datatype == 'ion':
            flux_omni = flux_omni*iGfact[0]
        if probe == '2' and datatype == 'ion':
            flux_omni = flux_omni*iGfact[1]
        if probe == '3' and datatype == 'ion':
            flux_omni = flux_omni*iGfact[2]
        if probe == '4' and datatype == 'ion':
            flux_omni = flux_omni*iGfact[3]

        store_data('mms'+probe+'_epd_feeps_'+data_rate+'_'+level+'_'+datatype+'_'+data_units+'_omni'+suffix, data={'x': tmpdata[0], 'y': flux_omni, 'v': energies})
        options('mms'+probe+'_epd_feeps_'+data_rate+'_'+level+'_'+datatype+'_'+data_units+'_omni'+suffix, 'spec', True)
        options('mms'+probe+'_epd_feeps_'+data_rate+'_'+level+'_'+datatype+'_'+data_units+'_omni'+suffix, 'ylog', True)
        options('mms'+probe+'_epd_feeps_'+data_rate+'_'+level+'_'+datatype+'_'+data_units+'_omni'+suffix, 'zlog', True)
        options('mms'+probe+'_epd_feeps_'+data_rate+'_'+level+'_'+datatype+'_'+data_units+'_omni'+suffix, 'Colormap', 'spedas')
        options('mms'+probe+'_epd_feeps_'+data_rate+'_'+level+'_'+datatype+'_'+data_units+'_omni'+suffix, 'ztitle', units_label)
        options('mms'+probe+'_epd_feeps_'+data_rate+'_'+level+'_'+datatype+'_'+data_units+'_omni'+suffix, 'ytitle', 'MMS' + str(probe) + ' ' + datatype)
        options('mms'+probe+'_epd_feeps_'+data_rate+'_'+level+'_'+datatype+'_'+data_units+'_omni'+suffix, 'ysubtitle', '[keV]')
        out_vars.append('mms'+probe+'_epd_feeps_'+data_rate+'_'+level+'_'+datatype+'_'+data_units+'_omni'+suffix)

    return out_vars









import logging
import numpy as np

from pyspedas.utilities.data_exists import data_exists
from pyspedas.cotrans.cotrans import cotrans
from pyspedas.analysis.tnormalize import tnormalize
from pyspedas.analysis.tcrossp import tcrossp
from pyspedas.analysis.tinterpol import tinterpol

from pytplot import get_data, store_data

logging.captureWarnings(True)
logging.basicConfig(format='%(asctime)s: %(message)s', datefmt='%d-%b-%y %H:%M:%S', level=logging.INFO)

def mms_pgs_xgse(mag_temp, pos_temp):
    """
    Generates the 'xgse' transformation matrix
    """
    mag_data = get_data(mag_temp)

    # xaxis of this system is X of the gse system. Z is mag field
    x_axis = np.zeros((len(mag_data.times), 3))
    x_axis[:, 0] = 1

    # create orthonormal basis set
    z_basis = tnormalize(mag_temp, return_data=True)
    y_basis = tcrossp(z_basis, x_axis, return_data=True)
    y_basis = tnormalize(y_basis, return_data=True)
    x_basis = tcrossp(y_basis, z_basis, return_data=True)

    return (x_basis, y_basis, z_basis)

def mms_pgs_phigeo(mag_temp, pos_temp):
    """
    Generates the 'phigeo' transformation matrix
    """
    pos_data = get_data(pos_temp)

    if pos_data is None:
        logging.error('Error with position data')
        return
        
    # transformation to generate other_dim dim for phigeo from thm_fac_matrix_make
    # All the conversions to polar and trig simplifies to this.
    # But the reason the conversion is why this is the conversion that is done, is lost on me.
    # The conversion swaps the x & y components of position, reflects over x=0,z=0 then projects into the xy plane
    pos_conv = np.stack((-pos_data.y[:, 1], pos_data.y[:, 0], np.zeros(len(pos_data.times))))
    pos_conv = np.transpose(pos_conv, [1, 0])
    store_data(pos_temp, data={'x': pos_data.times, 'y': pos_conv})

    # transform into GSE because the particles are in GSE
    cotrans(name_in=pos_temp, name_out=pos_temp, coord_in='gei', coord_out='gse')

    # create orthonormal basis set
    z_basis = tnormalize(mag_temp, return_data=True)
    x_basis = tcrossp(pos_temp, z_basis, return_data=True)
    x_basis = tnormalize(x_basis, return_data=True)
    y_basis = tcrossp(z_basis, x_basis, return_data=True)

    return (x_basis, y_basis, z_basis)

def mms_pgs_mphigeo(mag_temp, pos_temp):
    """
    Generates the 'mphigeo' transformation matrix
    """
    pos_data = get_data(pos_temp)

    if pos_data is None:
        logging.error('Error with position data')
        return

    # the following is heisted from the IDL version
    # transformation to generate other_dim dim for mphigeo from thm_fac_matrix_make
    # All the conversions to polar and trig simplifies to this.  
    # But the reason the conversion is why this is the conversion that is done, is lost on me.
    # The conversion swaps the x & y components of position, reflects over x=0,z=0 then projects into the xy plane
    pos_conv = np.stack((-pos_data.y[:, 1], pos_data.y[:, 0], np.zeros(len(pos_data.times))))
    pos_conv = np.transpose(pos_conv, [1, 0])
    store_data(pos_temp, data={'x': pos_data.times, 'y': pos_conv})

    # transform into GSE because the particles are in GSE
    cotrans(name_in=pos_temp, name_out=pos_temp, coord_in='gei', coord_out='gse')

    # create orthonormal basis set
    z_basis = tnormalize(mag_temp, return_data=True)
    x_basis = tcrossp(z_basis, pos_temp, return_data=True)
    x_basis = tnormalize(x_basis, return_data=True)
    y_basis = tcrossp(z_basis, x_basis, return_data=True)

    return (x_basis, y_basis, z_basis)


def mms_pgs_make_fac(times, mag_tvar_in, pos_tvar_in, fac_type='mphigeo'):
    """
    Generate the field aligned coordinate transformation matrix
    """

    if not data_exists(mag_tvar_in):
        logging.error('Magnetic field variable not found: "' + mag_tvar_in + '"; skipping field-aligned outputs')
        return

    if not data_exists(pos_tvar_in):
        logging.error('Position variable not found: "' + pos_tvar_in + '"; skipping field-aligned outputs')
        return

    valid_types = ['mphigeo', 'phigeo', 'xgse']

    if fac_type not in valid_types:
        logging.error('Invalid FAC type; valid types: "mphigeo", "phigeo", "xgse"')
        return

    cotrans(name_in=pos_tvar_in, name_out=pos_tvar_in, coord_in='gse', coord_out='gei')

    mag_temp = mag_tvar_in + '_pgs_temp'
    pos_temp = pos_tvar_in + '_pgs_temp'

    # first, we need to interpolate the magnetic field and position
    # variables to the time stamps of the particle data (times)
    tinterpol(mag_tvar_in, times, newname=mag_temp)
    tinterpol(pos_tvar_in, times, newname=pos_temp)

    if fac_type == 'mphigeo':
        basis = mms_pgs_mphigeo(mag_temp, pos_temp)
    elif fac_type == 'phigeo':
        basis = mms_pgs_phigeo(mag_temp, pos_temp)
    elif fac_type == 'xgse':
        basis = mms_pgs_xgse(mag_temp, pos_temp)

    fac_output = np.zeros((len(times), 3, 3))
    fac_output[:, 0, :] = basis[0]
    fac_output[:, 1, :] = basis[1]
    fac_output[:, 2, :] = basis[2]

    return fac_output


from pyspedas.mms.mms_load_data import mms_load_data
from pyspedas.mms.fpi.mms_fpi_set_metadata import mms_fpi_set_metadata
from pyspedas.mms.fpi.mms_load_fpi_calc_pad import mms_load_fpi_calc_pad
from pyspedas.mms.fpi.mms_fpi_make_compressionlossbars import mms_fpi_make_compressionlossbars
from pyspedas.mms.fpi.mms_fpi_make_errorflagbars import mms_fpi_make_errorflagbars
from pyspedas.mms.print_vars import print_vars
from pyspedas.mms.mms_config import CONFIG
from pytplot import tplot_rename, del_data

@print_vars
def mms_load_fpi(trange=['2015-10-16', '2015-10-17'], probe='1', data_rate='fast',
    level='l2', datatype='*', varformat=None, varnames=[], suffix='',
    get_support_data=False, time_clip=False, no_update=False, center_measurement=False,
    available=False, notplot=False, latest_version=False, major_version=False, 
    min_version=None, cdf_version=None, spdf=False, always_prompt=False):
    """
    This function loads FPI data into tplot variables
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe : str or list of str
            list of probes, valid values for MMS probes are ['1','2','3','4']. 

        data_rate : str or list of str
            instrument data rates for FPI include 'brst', 'fast'. The
            default is 'srvy'.

        level : str
            indicates level of data processing. the default if no level is specified is 'l2'

        datatype : str or list of str
            Valid datatypes for FPI are:
             'des-moms', 'dis-moms' (default)
             'des-dist', 'dis-dist'

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        time_clip: bool
            Data will be clipped to the exact trange specified by the trange keyword.
            
        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        center_measurement: bool
            If True, the CDF epoch variables are time-shifted to the middle
            of the accumulation interval by their DELTA_PLUS_VAR and
            DELTA_MINUS_VAR variable attributes

        notplot: bool
            If True, then data are returned in a hash table instead of 
            being stored in tplot variables (useful for debugging, and
            access to multi-dimensional data products)

        available: bool
            If True, simply return the available data files (without downloading)
            for the requested paramters

        no_update: bool
            Set this flag to preserve the original data. if not set and newer 
            data is found the existing data will be overwritten

        cdf_version: str
            Specify a specific CDF version # to load (e.g., cdf_version='4.3.0')

        min_version: str
            Specify a minimum CDF version # to load

        latest_version: bool
            Only grab the latest CDF version in the requested time interval

        major_version: bool
            Only open the latest major CDF version (e.g., X in vX.Y.Z) in the requested time interval

        always_prompt: bool
            Set this keyword to always prompt for the user's username and password;
            useful if you accidently save an incorrect password, or if your SDC password has changed

        spdf: bool
            If True, download the data from the SPDF instead of the SDC
            
    Returns:
        List of tplot variables created.

    """

    # different datatypes for burst mode files
    if data_rate.lower() == 'brst':
        if isinstance(datatype, str):
            if (datatype == '*' or datatype == '') and level.lower() != 'ql':
                datatype = ['des-dist', 'dis-dist', 'dis-moms', 'des-moms']
    else:
        if isinstance(datatype, str):
            if (datatype == '*' or datatype == '') and level.lower() == 'ql':
                datatype = ['des', 'dis']
            if (datatype == '*' or datatype == '') and level.lower() != 'ql':
                datatype = ['des-dist', 'dis-dist', 'dis-moms', 'des-moms']

    # kludge for level = 'sitl' -> datatype shouldn't be defined for sitl data.
    if level.lower() == 'sitl' or level.lower() == 'trig':
        datatype = ''

    if varformat is None and varnames == []:
        get_support_data = True

    tvars = mms_load_data(trange=trange, probe=probe, data_rate=data_rate, level=level, instrument='fpi',
            datatype=datatype, varformat=varformat, varnames=varnames, suffix=suffix, get_support_data=get_support_data,
            time_clip=time_clip, no_update=no_update, center_measurement=center_measurement, available=available, 
            notplot=notplot, latest_version=latest_version, major_version=major_version, min_version=min_version, 
            cdf_version=cdf_version, spdf=spdf, always_prompt=always_prompt)
    
    if tvars is None or available or notplot or CONFIG['download_only']:
        return tvars

    mms_fpi_set_metadata(probe, data_rate, datatype, level, suffix=suffix)

    if not isinstance(probe, list): probe = [probe]
    if not isinstance(data_rate, list): data_rate = [data_rate]
    if not isinstance(datatype, list): datatype = [datatype]
    if not isinstance(level, list): level = [level]

    # the following kludge is due to the errorflags variable in the dist and moments files having the
    # same variable name, so loading d?s-dist and d?s-moms files at the same time will overwrite
    # one of the vars containing errorflags
    if 'des-dist' in datatype and 'des-moms' in datatype:
        # delete the old vars first
        del_data('*_des_errorflags_*')
        del_data('*_des_compressionloss_*')
        # load the bars with a suffix applied
        tplotnames_errflags_emom = mms_load_data(trange=trange, probe=probe, data_rate=data_rate, level=level, instrument='fpi',
                              datatype='des-moms', varformat='*errorflags*|*compressionloss*', varnames=varnames,
                              suffix=suffix+'_moms', get_support_data=0,
                              time_clip=time_clip, no_update=no_update, center_measurement=center_measurement,
                              available=available,
                              notplot=notplot, latest_version=latest_version, major_version=major_version,
                              min_version=min_version,
                              cdf_version=cdf_version, spdf=spdf, always_prompt=always_prompt)
        tplotnames_errflags_edist = mms_load_data(trange=trange, probe=probe, data_rate=data_rate, level=level, instrument='fpi',
                              datatype='des-dist', varformat='*errorflags*|*compressionloss*', varnames=varnames,
                              suffix=suffix+'_dist', get_support_data=0,
                              time_clip=time_clip, no_update=no_update, center_measurement=center_measurement,
                              available=available,
                              notplot=notplot, latest_version=latest_version, major_version=major_version,
                              min_version=min_version,
                              cdf_version=cdf_version, spdf=spdf, always_prompt=always_prompt)
        tvars.extend(tplotnames_errflags_emom)
        tvars.extend(tplotnames_errflags_edist)
    else:
        # user didn't request both dist and moments, so no variables should have been clobbered
        # but we still need to append _dist, _moms to the end of the names
        for prb in probe:
            for drate in data_rate:
                this_probe = str(prb)
                if 'des-dist' in datatype:
                    tplot_rename('mms'+this_probe+'_des_errorflags_'+drate+suffix, 'mms'+this_probe+'_des_errorflags_'+drate+suffix+'_dist')
                    tplot_rename('mms'+this_probe+'_des_compressionloss_'+drate+suffix, 'mms'+this_probe+'_des_compressionloss_'+drate+suffix+'_dist')
                    tvars.append('mms'+this_probe+'_des_errorflags_'+drate+suffix+'_dist')
                    tvars.append('mms'+this_probe+'_des_compressionloss_'+drate+suffix+'_dist')
                if 'des-moms' in datatype or 'des' in datatype:
                    tplot_rename('mms'+this_probe+'_des_errorflags_'+drate+suffix, 'mms'+this_probe+'_des_errorflags_'+drate+suffix+'_moms')
                    tplot_rename('mms'+this_probe+'_des_compressionloss_'+drate+suffix, 'mms'+this_probe+'_des_compressionloss_'+drate+suffix+'_moms')
                    tvars.append('mms'+this_probe+'_des_errorflags_'+drate+suffix+'_moms')
                    tvars.append('mms'+this_probe+'_des_compressionloss_'+drate+suffix+'_moms')

    # same kludge as above, but for ions
    if 'dis-dist' in datatype and 'dis-moms' in datatype:
        # delete the old vars first
        del_data('*_dis_errorflags_*')
        del_data('*_dis_compressionloss_*')
        # load the bars with a suffix applied
        tplotnames_errflags_imom = mms_load_data(trange=trange, probe=probe, data_rate=data_rate, level=level, instrument='fpi',
                              datatype='dis-moms', varformat='*errorflags*|*compressionloss*', varnames=varnames,
                              suffix=suffix+'_moms', get_support_data=0,
                              time_clip=time_clip, no_update=no_update, center_measurement=center_measurement,
                              available=available,
                              notplot=notplot, latest_version=latest_version, major_version=major_version,
                              min_version=min_version,
                              cdf_version=cdf_version, spdf=spdf, always_prompt=always_prompt)
        tplotnames_errflags_idist = mms_load_data(trange=trange, probe=probe, data_rate=data_rate, level=level, instrument='fpi',
                              datatype='dis-dist', varformat='*errorflags*|*compressionloss*', varnames=varnames,
                              suffix=suffix+'_dist', get_support_data=0,
                              time_clip=time_clip, no_update=no_update, center_measurement=center_measurement,
                              available=available,
                              notplot=notplot, latest_version=latest_version, major_version=major_version,
                              min_version=min_version,
                              cdf_version=cdf_version, spdf=spdf, always_prompt=always_prompt)
        tvars.extend(tplotnames_errflags_imom)
        tvars.extend(tplotnames_errflags_idist)
    else:
        # user didn't request both dist and moments, so no variables should have been clobbered
        # but we still need to append _dist, _moms to the end of the names
        for prb in probe:
            for drate in data_rate:
                this_probe = str(prb)
                if 'dis-dist' in datatype:
                    tplot_rename('mms'+this_probe+'_dis_errorflags_'+drate+suffix, 'mms'+this_probe+'_dis_errorflags_'+drate+suffix+'_dist')
                    tplot_rename('mms'+this_probe+'_dis_compressionloss_'+drate+suffix, 'mms'+this_probe+'_dis_compressionloss_'+drate+suffix+'_dist')
                    tvars.append('mms'+this_probe+'_dis_errorflags_'+drate+suffix+'_dist')
                    tvars.append('mms'+this_probe+'_dis_compressionloss_'+drate+suffix+'_dist')
                if 'dis-moms' in datatype or 'dis' in datatype:
                    tplot_rename('mms'+this_probe+'_dis_errorflags_'+drate+suffix, 'mms'+this_probe+'_dis_errorflags_'+drate+suffix+'_moms')
                    tplot_rename('mms'+this_probe+'_dis_compressionloss_'+drate+suffix, 'mms'+this_probe+'_dis_compressionloss_'+drate+suffix+'_moms')
                    tvars.append('mms'+this_probe+'_dis_errorflags_'+drate+suffix+'_moms')
                    tvars.append('mms'+this_probe+'_dis_compressionloss_'+drate+suffix+'_moms')

    for prb in probe:
        for drate in data_rate:
            for dtype in datatype:
                for lvl in level:
                    out_var = mms_load_fpi_calc_pad(probe=prb, level=lvl, datatype=dtype, data_rate=drate, suffix=suffix, autoscale=True)
                    if out_var:
                        tvars.extend(out_var)

                    # create the error bars
                    # moments
                    if dtype == 'des-moms':
                        des_moms_eflags = mms_fpi_make_errorflagbars('mms' + str(prb) + '_des_errorflags_' + drate + '_moms', level=lvl)
                        if des_moms_eflags is not None:
                            tvars.extend(des_moms_eflags)

                    if dtype == 'dis-moms':
                        dis_moms_eflags = mms_fpi_make_errorflagbars('mms' + str(prb) + '_dis_errorflags_' + drate + '_moms', level=lvl)
                        if dis_moms_eflags is not None:
                            tvars.extend(dis_moms_eflags)

                    # distributions
                    if dtype == 'des-dist':
                        des_dist_eflags = mms_fpi_make_errorflagbars('mms' + str(prb) + '_des_errorflags_' + drate + '_dist', level=lvl)
                        if des_dist_eflags is not None:
                            tvars.extend(des_dist_eflags)

                    if dtype == 'dis-dist':
                        dis_dist_eflags = mms_fpi_make_errorflagbars('mms' + str(prb) + '_dis_errorflags_' + drate + '_dist', level=lvl)
                        if dis_dist_eflags is not None:
                            tvars.extend(dis_dist_eflags)

                    if drate == 'brst':
                        des_moms_comp = mms_fpi_make_compressionlossbars('mms' + str(prb) + '_des_compressionloss_' + drate + '_moms')
                        dis_moms_comp = mms_fpi_make_compressionlossbars('mms' + str(prb) + '_dis_compressionloss_' + drate + '_moms')
                        des_dist_comp = mms_fpi_make_compressionlossbars('mms' + str(prb) + '_des_compressionloss_' + drate + '_dist')
                        dis_dist_comp = mms_fpi_make_compressionlossbars('mms' + str(prb) + '_dis_compressionloss_' + drate + '_dist')

                        if des_moms_comp is not None:
                            tvars.extend(des_moms_comp)
                        if dis_moms_comp is not None:
                            tvars.extend(dis_moms_comp)
                        if des_dist_comp is not None:
                            tvars.extend(des_dist_comp)
                        if dis_dist_comp is not None:
                            tvars.extend(dis_dist_comp)
    return tvars



from copy import deepcopy
import numpy as np

from .slice2d_intrange import slice2d_intrange
from .slice2d_get_sphere import slice2d_get_sphere
from .slice2d_checkbins import slice2d_checkbins
from .slice2d_collate import slice2d_collate
from .slice2d_get_ebounds import slice2d_get_ebounds


def slice2d_get_data(dists, trange=None, energy=False, erange=None):
    """
    Returns an array of averaged data along with the corresponding
    bin centers and widths in spherical coordinates. This routine
    will apply energy range constraints and count thresholds.

    Input
    ------
        dists: list of dicts
            List of 3D particle data structures

    Parameters
    -----------
        trange: list
            Time range

        energy: bool
            Flag to get energy bins instead of velocity bins for radial distance

        erange: list
            Two element array specifying min/max energies to be used

    Returns
    --------
        Dictionary containing data, bin centers and widths:
            data: N element array containing averaged particle data
            rad: N element array of bin centers along r (eV or km/s)
            phi: N element array of bin centers along phi
            theta: N element array of bin centers along theta
            dr: N element array of bin widths along r (eV or km/s)
            dp: N element array of bin widths along phi
            dt: N element array of bin widths along theta
    """
    # Get indexes of data structures in requested time window
    times_idx = slice2d_intrange(dists, trange)

    if len(times_idx) == 0:
        print('No data in the time range')
        return

    weight = np.zeros(dists[0]['bins'][:, :, :].shape)
    data = np.zeros(dists[0]['data'][:, :, :].shape)
    out = None
    sphere = None
    last = None

    for time in times_idx:
        dist = dists[time]

        # if this sample's angle or energy bins or mass differ from last then
        # collate any aggregated data and continue
        if not slice2d_checkbins(dist, last):
            out = slice2d_collate(data, weight, sphere, previous_out=out)
            weight = np.zeros(dists[0]['bins'][:, :, :].shape)
            data = np.zeros(dists[0]['data'][:, :, :].shape)

        # copy current data for comparison in next iteration
        last = deepcopy(dist)

        # find active, valid bins
        bins = (dist['bins'] != 0)*1 & (np.isfinite(dist['data'])*1)

        # find bins within energy limits
        if erange is not None:
            n = dist['energy'].shape[0]
            energies = slice2d_get_ebounds(dist)
            ecenters = (energies[0:n, :, :]+energies[1:n+1, :, :])/2.0
            bins = bins & ((ecenters >= erange[0]) & (ecenters <= erange[1]))

        # Get data & bin boundaries
        # Coordinates will only be calculated/copied if the arrays are not 
        # in existence, otherwise this just copies the data and bins arrays.
        sphere = slice2d_get_sphere(dist, energy=energy)

        # sum of counts over current set of samples
        data = data + dist['data']

        # keep track of valid bins within range. this array will be used later
        # to average bins and discard any that are out of range or invalid
        weight = weight + bins

    # collate any remaining data
    out = slice2d_collate(data, weight, sphere, previous_out=out)

    return out




import os
import unittest
from pyspedas.utilities.data_exists import data_exists

import pyspedas

class LoadTestCases(unittest.TestCase):
    def test_downloadonly(self):
        files = pyspedas.fast.acb(trange=['1999-09-22', '1999-09-23'], time_clip=True, level='k0', downloadonly=True)
        self.assertTrue(os.path.exists(files[0]))

    def test_load_dcb_data(self):
        dcb_vars = pyspedas.fast.dcb(trange=['1999-09-22', '1999-09-23'], time_clip=True, level='k0')
        self.assertTrue(data_exists('BX'))
        self.assertTrue(data_exists('BY'))
        self.assertTrue(data_exists('BZ'))

    def test_load_acb_data(self):
        dcb_vars = pyspedas.fast.acb(trange=['1999-09-22', '1999-09-23'], time_clip=True, level='k0')
        self.assertTrue(data_exists('HF_PWR'))
        self.assertTrue(data_exists('HF_E_SPEC'))

    # as of 8 Feb 2020, these data fail to load in IDL and Python
    # def test_load_esa_data(self):
    #     esa_vars = pyspedas.fast.esa()
    #     self.assertTrue(data_exists(''))

    def test_load_teams_data(self):
        teams_vars = pyspedas.fast.teams()
        self.assertTrue(data_exists('H+'))
        self.assertTrue(data_exists('O+'))
        self.assertTrue(data_exists('He+'))

if __name__ == '__main__':
    unittest.main()


"""
;    To transform time series data from one coordinate to another. The supported coordinate 
;     systems are: SGA, SGI, DSI, J2000. Further transformation from J2000 to the 
;     geophysical coordinates (GEI, GSE, etc) can be processed by "cotrans()", which can be
      imported by (from pyspedas.cotrans.cotrans import cotrans). 
;     
;     The actual transformation is done by helper functions, such as sga2sgi(), sgi2dsi(), 
;     and dsi2j2000().
            
Main routine for coordinate transformation is erg_cotrans().
"""

from pyspedas import tnames
from pytplot import tplot_copy

from .dsi2j2000 import dsi2j2000
from .sga2sgi import sga2sgi
from .sgi2dsi import sgi2dsi


def erg_replace_coord_suffix(in_name=None, out_coord=None):
    nms = in_name.split('_')
    nms[-1] = out_coord
    return '_'.join(nms)


def erg_coord_trans(in_name=None,
                    out_name=None,
                    in_coord=None,
                    out_coord=None,
                    noload=False):

    if in_coord == out_coord:
        tplot_copy(in_name, out_name)
        return

    # From SGA
    if in_coord == 'sga':
        if out_coord == 'sgi':  # sga --> sgi
            sga2sgi(name_in=in_name, name_out=out_name, noload=noload)
        elif out_coord == 'dsi':  # sga --> sgi --> dsi
            name_sgi = erg_replace_coord_suffix(
                in_name=in_name, out_coord='sgi')
            sga2sgi(name_in=in_name, name_out=name_sgi, noload=noload)
            sgi2dsi(name_in=name_sgi, name_out=out_name, noload=noload)
        elif out_coord == 'j2000':  # sga --> sgi --> dsi --> j2000
            name_sgi = erg_replace_coord_suffix(
                in_name=in_name, out_coord='sgi')
            sga2sgi(name_in=in_name, name_out=name_sgi, noload=noload)
            name_dsi = erg_replace_coord_suffix(
                in_name=name_sgi, out_coord='dsi')
            sgi2dsi(name_in=name_sgi, name_out=name_dsi, noload=noload)
            dsi2j2000(name_in=name_dsi, name_out=out_name, noload=noload)

    # From SGI
    if in_coord == 'sgi':
        if out_coord == 'sga':  # sgi --> sga
            sga2sgi(name_in=in_name, name_out=out_name,
                    SGI2SGA=True, noload=noload)
        elif out_coord == 'dsi':  # sgi --> dsi
            sgi2dsi(name_in=in_name, name_out=out_name, noload=noload)
        elif out_coord == 'j2000':  # sgi --> dsi --> j2000
            name_dsi = erg_replace_coord_suffix(
                in_name=in_name, out_coord='dsi')
            sgi2dsi(name_in=in_name, name_out=name_dsi, noload=noload)
            dsi2j2000(name_in=name_dsi, name_out=out_name, noload=noload)

    # From DSI
    elif in_coord == 'dsi':

        if out_coord == 'sga':  # dsi --> sgi --> sga
            name_sgi = erg_replace_coord_suffix(
                in_name=in_name, out_coord='sgi')
            sgi2dsi(name_in=in_name, name_out=name_sgi,
                    DSI2SGI=True, noload=noload)
            sga2sgi(name_in=name_sgi, name_out=out_name,
                    SGI2SGA=True, noload=noload)
        elif out_coord == 'sgi':  # dsi --> sgi
            sgi2dsi(name_in=in_name, name_out=out_name,
                    DSI2SGI=True, noload=noload)
        elif out_coord == 'j2000':  # dsi --> j2000
            dsi2j2000(name_in=in_name, name_out=out_name, noload=noload)

    # From J2000
    elif in_coord == 'j2000':
        if out_coord == 'sga':  # j2000 --> dsi --> sgi --> sga
            name_dsi = erg_replace_coord_suffix(
                in_name=in_name, out_coord='dsi')
            dsi2j2000(name_in=in_name, name_out=name_dsi,
                      J20002DSI=True, noload=noload)
            name_sgi = erg_replace_coord_suffix(
                in_name=name_dsi, out_coord='sgi')
            sgi2dsi(name_in=name_dsi, name_out=name_sgi,
                    DSI2SGI=True, noload=noload)
            sga2sgi(name_in=name_sgi, name_out=out_name,
                    SGI2SGA=True, noload=noload)
        elif out_coord == 'sgi':  # j2000 --> dsi --> sgi
            name_dsi = erg_replace_coord_suffix(
                in_name=in_name, out_coord='dsi')
            dsi2j2000(name_in=in_name, name_out=name_dsi,
                      J20002DSI=True, noload=noload)
            sgi2dsi(name_in=name_dsi, name_out=out_name,
                    DSI2SGI=True, noload=noload)
        elif out_coord == 'dsi':  # j2000 --> dsi
            dsi2j2000(name_in=in_name, name_out=out_name,
                      J20002DSI=True, noload=noload)

# ;;;; Main routine for coordinate transformation ;;;;;


def erg_cotrans(in_name='',
                out_name='',
                in_coord='',
                out_coord='',
                noload=False):
    """
    Parameters:

        in_name : str
            name of input tplot variable to be transformed

        out_name : str
            Name of output tplot variable in which the transformed data are stored. 
             If not explicitly provided, out_name is automatically generated from 
             "in_name" by replacing the coordinate name with a new one. 
             For example, if you runs erg_cotrans(in_name='xxxx_sgi', out_coord='sga')
             then the result is stored in a newly created tplot variable 'xxxx_sga'.

        in_coord : str
            'sga', 'sgi', 'dsi', or 'j2000'
            Set to explicitly give the coordinate system name for the input variable.
            If not given, this routine tries to guess it from the name of the input
            tplot variable (in_name).

        out_coord : str
            'sga', 'sgi', 'dsi', or 'j2000'
            Set to explicitly give the coordinate system name for the output variable.
            If not given, the coordinate system name is obtained from the variable name
            as done for in_coord.

    Returns:
        None

    """
    valid_suffixes = ['sga', 'sgi', 'dsi', 'j2000']

    in_names = tnames(in_name)
    if len(in_names) == 0:
        print('in_name is not match in stored Tplot Variables, return')
        return

    for input_name in in_names:
        if in_coord not in valid_suffixes:
            if (input_name.split('_')[-1] in valid_suffixes):
                in_suf = input_name.split('_')[-1]
            else:
                print(
                    f'Cannot get a valid coord. suffix of input variable for {input_name}.')
                continue
        else:
            in_suf = in_coord
        if out_coord not in valid_suffixes:
            if out_name.split('_')[-1] in valid_suffixes:
                out_suf = out_name.split('_')[-1]
            else:
                print(
                    f'Cannot get a valid coord. suffix of output variable for {input_name}.')
                continue
        else:
            out_suf = out_coord
        if out_name == '':
            out_name_temp = erg_replace_coord_suffix(
                in_name=input_name, out_coord=out_suf)
            erg_coord_trans(in_name=input_name, out_name=out_name_temp,
                            in_coord=in_suf, out_coord=out_suf, noload=noload)
        else:
            erg_coord_trans(in_name=input_name, out_name=out_name,
                            in_coord=in_suf, out_coord=out_suf, noload=noload)



import os
import fnmatch
import glob
import re
import logging
import pandas as pd

from pyspedas import time_double
from pyspedas.mms.mms_config import CONFIG

def mms_get_local_state_files(probe='1', level='def', filetype='eph', trange=None):
    """
    Search for local state MMS files in case a list cannot be retrieved from the
    remote server.  Returns a sorted list of file paths.
    
    Parameters:
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe: str
            probe #, e.g., '4' for MMS4
    
        level: str
            state data level; either 'def' (for definitive) or 'pred' (for predicted)

        filetype: str
            state file type, e.g. 'eph' (for ephemeris) or 'att' (for attitude)

    Returns:
        List of local files found matching the input parameters.

    """
    if trange is None:
        logging.info('No trange specified in mms_get_local_state_files')
        return

    # directory and file name search patterns
    # For now
    # -all ancillary data is in one directory:
    #       mms\ancillary
    # -assume file names are of the form:
    #   SPACECRAFT_FILETYPE_startDate_endDate.version
    #   where SPACECRAFT is [MMS1, MMS2, MMS3, MMS4] in uppercase
    #   and FILETYPE is either DEFATT, PREDATT, DEFEPH, PREDEPH in uppercase
    #   and start/endDate is YYYYDOY
    #   and version is Vnn (.V00, .V01, etc..)
    dir_pattern = os.sep.join([CONFIG['local_data_dir'], 'ancillary', 'mms'+probe, level+filetype])
    file_pattern = 'MMS'+probe+'_'+level.upper()+filetype.upper()+'_'+'???????_???????.V??'

    files_in_trange = []
    out_files = []

    files = glob.glob(os.sep.join([dir_pattern, file_pattern]))

    # find the files within the trange
    file_regex = re.compile(os.sep.join([dir_pattern, 'MMS'+probe+'_'+level.upper()+filetype.upper()+'_([0-9]{7})_([0-9]{7}).V[0-9]{2}']))
    for file in files:
        time_match = file_regex.match(file)
        if time_match != None:
            start_time = pd.to_datetime(time_match.group(1), format='%Y%j').timestamp()
            end_time = pd.to_datetime(time_match.group(2), format='%Y%j').timestamp()
            if start_time < time_double(trange[1]) and end_time >= time_double(trange[0]):
                files_in_trange.append(file)

    # ensure only the latest version of each file is loaded
    for file in files_in_trange:
        this_file = file[0:-3] + 'V??'
        versions = fnmatch.filter(files_in_trange, this_file)
        if len(versions) > 1:
            out_files.append(sorted(versions)[-1]) # only grab the latest version
        else:
            out_files.append(versions[0])

    return list(set(out_files))


from copy import deepcopy
from time import time
import numpy as np


def slice2d_geo(data, resolution, r, phi, theta, dr, dp, dt, orient_matrix=None, rotation_matrix=None,
                custom_matrix=None, msg_prefix='', shift=None):
    """
    Produces slices showing each bin's boundaries by assigning
    each bin's value to all points on the slice plane that
    fall within that bin's boundaries.
    """
    n = float(resolution)
    n_int = int(n)
    rd = 180/np.pi
    tolerance = 5e-7 # to account for rounding errors
    # for progress messages
    previous_time = time()

    vrange = [-np.max(np.abs(r)), np.max(np.abs(r))]
    dr_range = [-np.max(dr), np.max(np.abs(dr))]
    xgrid = np.linspace(vrange[0]+dr_range[0], vrange[1]+dr_range[1], n_int)
    ygrid = np.linspace(vrange[0]+dr_range[0], vrange[1]+dr_range[1], n_int)
    z = 0

    uvals = np.zeros((n_int**2, 3))
    xvals = np.outer(xgrid, np.ones(n_int)).flatten(order='F')
    yvals = np.outer(np.ones(n_int), ygrid).flatten(order='F')
    zvals = np.zeros(n_int**2)
    uvals[:, 0] = xvals
    uvals[:, 1] = yvals
    uvals[:, 2] = zvals

    m = deepcopy(orient_matrix)
    # rotate slice coordinates to desired location
    if custom_matrix is not None:
        if rotation_matrix is not None:
            m = (custom_matrix @ rotation_matrix) @ m
        else:
            m = custom_matrix @ m
    else:
        if rotation_matrix is not None:
            m = rotation_matrix @ m
    uvals = uvals @ m.T

    na = 0
    num_points = len(data)
    out = np.zeros(n_int**2)
    weight = np.zeros(n_int**2)

    # loop over slice plans (if averaging over angle)
    for j in range(-1, na):
        ut = deepcopy(uvals)

        # Convert transformed slice coordinates to spherical
        pcoords = rd*np.arctan2(ut[:, 1], ut[:, 0])  # phi
        tcoords = rd*np.arctan2(ut[:, 2], np.sqrt(ut[:, 0]**2 + ut[:, 1]**2))  # theta
        rcoords = np.sqrt(ut[:, 0]**2 + ut[:, 1]**2 + ut[:, 2]**2)  # r

        # Loop over bins to determine what region each bin covers on the slice plane
        for i in range(0, num_points):
            if data[i] == 0:
                continue

            # theta
            tlim = np.array([theta[i]-0.5*dt[i], theta[i]+0.5*dt[i]])

            # account for rounding errors
            # this is particularly important is slice plane is at zero elevation
            tr = np.argwhere(np.abs(tlim - np.round(tlim)) < tolerance).flatten()
            if len(tr) > 0:
               tlim[tr] = np.round(tlim[tr])

            thetas =(tcoords > tlim[0]) & (tcoords <= tlim[1])

            if len(thetas) == 0:
                continue

            if np.sum(thetas) < 1:
                continue
            
            # phi
            plim = np.array([phi[i]-0.5*dp[i], phi[i]+0.5*dp[i]])

            # keep limits within [-180, 180]
            over = np.argwhere(plim > 180).flatten()
            under = np.argwhere(plim < -180).flatten()
            if len(over) > 0:
                plim[over] += -360.0
            if len(under) > 0:
                plim[under] += 360.0
            
            # account for rounding errors
            pr = np.argwhere(np.abs(plim - np.round(plim)) < tolerance).flatten()
            if len(pr) > 0:
               plim[pr] = np.round(plim[pr])
            
            # determine which region ( p0->p1 or p1->p0) the bin spans
            if plim[0] > plim[1]:
                phis = (pcoords > plim[0]) | (pcoords <= plim[1])
            else:
                phis = (pcoords > plim[0]) & (pcoords <= plim[1])
            
            if len(phis) == 0:
                continue

            if np.sum(phis) < 1:
                continue

            # R (velocity/energy)
            rdata = (rcoords >= r[i]-0.5*dr[i]) & (rcoords < r[i]+0.5*dr[i])

            bidx = np.argwhere(phis & thetas & rdata).flatten()

            if len(bidx) > 1:
                weight[bidx] = weight[bidx] + 1
                out[bidx] = out[bidx] + data[i]

            # output progress messages every 6 seconds
            if (time() - previous_time) > 6:
                if na == 0:
                    num_angles = 1
                else:
                    num_angles = na
                msg = msg_prefix + str(int(100*((j+1)*num_points + i)/(num_angles*num_points))) + '% complete'
                print(msg)
                previous_time = time()

    # average areas where bins overlapped
    adj = np.argwhere(weight == 0)
    if len(adj) > 0:
        weight[adj] = 1
    out = out / weight

    out = out.reshape((n_int, n_int), order='F')

    if shift is not None:
        xgrid -= shift[0]
        ygrid -= shift[1]

    return {'data': out, 'xgrid': xgrid, 'ygrid': ygrid}



import unittest
import pyspedas
from pyspedas.utilities.data_exists import data_exists
from pyspedas.cotrans.fac_matrix_make import fac_matrix_make
from pyspedas import tinterpol
from pytplot import get_data, store_data
from pyspedas import twavpol


class WavpolTestCases(unittest.TestCase):
    def test_brst(self):
        """
        See the MMS wave polarization notebook for details/notes on these calculations
        """
        trange = ['2015-10-16/13:05:40', '2015-10-16/13:07:25']  # Burch et al., Science event
        sc = '4'
        scm_data_rate = 'brst'
        scm_datatype = 'scb'
        pyspedas.mms.scm(probe=sc, datatype=scm_datatype, level='l2', trange=trange, data_rate=scm_data_rate)
        mms_scm_name = 'mms' + sc + '_scm_acb_gse_' + scm_datatype + '_' + scm_data_rate + '_l2'
        fgm_data_rate = 'srvy'
        pyspedas.mms.fgm(probe=sc, trange=trange, data_rate=fgm_data_rate)
        mms_fgm_name = 'mms' + sc + '_fgm_b_gse_' + fgm_data_rate + '_l2_bvec'
        fac_matrix_make(mms_fgm_name, other_dim='xgse', newname=mms_fgm_name + '_fac_mat')
        tinterpol('mms4_fgm_b_gse_srvy_l2_bvec_fac_mat', mms_scm_name)
        fac_mat = get_data('mms4_fgm_b_gse_srvy_l2_bvec_fac_mat-itrp')
        scm_data = get_data(mms_scm_name)
        scm_fac = [fac_mat.y[idx, :, :] @ scm_data.y[idx, :] for idx in range(0, len(scm_data.y[:, 0]))]
        store_data(mms_scm_name + '_fac', data={'x': scm_data.times, 'y': scm_fac})
        # number of points for FFT
        nopfft_input = 8192  # 1024
        # number of points for shifting between 2 FFT
        steplength_input = nopfft_input / 2
        # number of bins for frequency averaging
        bin_freq_input = 3
        twavpol(mms_scm_name + '_fac', nopfft=nopfft_input, steplength=steplength_input, bin_freq=bin_freq_input)
        self.assertTrue(data_exists(mms_scm_name + '_fac'))
        self.assertTrue(data_exists('mms4_scm_acb_gse_scb_brst_l2_fac_powspec'))
        self.assertTrue(data_exists('mms4_scm_acb_gse_scb_brst_l2_fac_degpol'))
        self.assertTrue(data_exists('mms4_scm_acb_gse_scb_brst_l2_fac_waveangle'))
        self.assertTrue(data_exists('mms4_scm_acb_gse_scb_brst_l2_fac_elliptict'))
        self.assertTrue(data_exists('mms4_scm_acb_gse_scb_brst_l2_fac_helict'))


if __name__ == '__main__':
    unittest.main()



import numpy as np
from .slice2d_get_support import slice2d_get_support


def slice2d_custom_rotation(custom_rotation=None,
                            trange=None,
                            vectors=None,
                            bfield=None,
                            vbulk=None,
                            sunvec=None,
                            determ_tolerance=1e-6):
    """
    Retrieve a user-provided rotation matrix and apply to data as needed.
    """

    if custom_rotation is None:
        matrix = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        return {'matrix': matrix,
                'vectors': vectors,
                'vbulk': vbulk,
                'bfield': bfield,
                'sunvec': sunvec}

    matrix = slice2d_get_support(custom_rotation, trange, matrix=True)

    print('Applying custom rotation')

    # Transform particle and support vectors
    if vectors is not None:
        transformed = np.zeros(vectors.shape)
        num_vectors = vectors.shape[0]
        for vector_idx in range(num_vectors):
            transformed[vector_idx] = matrix.T @ vectors[vector_idx, :]
        vectors = transformed
    if vbulk is not None:
        vbulk = matrix @ vbulk
    if bfield is not None:
        bfield = matrix @ bfield
    if sunvec is not None:
        sunvec = matrix @ sunvec

    return {'matrix': matrix,
            'vectors': vectors,
            'vbulk': vbulk,
            'bfield': bfield,
            'sunvec': sunvec}


from .version import version
from .utilities.data_exists import data_exists
from .utilities.tnames import tnames
from .utilities.time_string import time_string, time_datetime
from .utilities.time_double import time_float, time_double
from .utilities.tcopy import tcopy
from .utilities.tkm2re import tkm2re

from .analysis.avg_data import avg_data
from .analysis.clean_spikes import clean_spikes
from .analysis.deriv_data import deriv_data
from .analysis.dpwrspc import dpwrspc
from .analysis.subtract_average import subtract_average
from .analysis.subtract_median import subtract_median
from .analysis.time_clip import time_clip
from .analysis.tdeflag import tdeflag
from .analysis.tdpwrspc import tdpwrspc
from .analysis.tinterpol import tinterpol
from .analysis.tnormalize import tnormalize
from .analysis.tdotp import tdotp
from .analysis.tcrossp import tcrossp
from .analysis.tsmooth import tsmooth
from .analysis.yclip import yclip
from .analysis.twavpol import twavpol
from pytplot import cdf_to_tplot

from .cotrans.cotrans import cotrans
from .cotrans.cotrans_get_coord import cotrans_get_coord
from .cotrans.cotrans_set_coord import cotrans_set_coord

from .mms import mms_load_mec, mms_load_fgm, mms_load_scm, mms_load_edi, mms_load_edp, mms_load_eis, mms_load_feeps, \
    mms_load_hpca, mms_load_fpi, mms_load_aspoc, mms_load_dsp, mms_load_fsm, mms_load_state
from .mms.feeps.mms_feeps_pad import mms_feeps_pad
from .mms.feeps.mms_feeps_gpd import mms_feeps_gpd
from .mms.eis.mms_eis_pad import mms_eis_pad
from .mms.hpca.mms_hpca_calc_anodes import mms_hpca_calc_anodes
from .mms.hpca.mms_hpca_spin_sum import mms_hpca_spin_sum

from .maven import maven_load

from . import erg
from . import ulysses
from . import mica
from . import goes
from . import themis
from . import omni
from . import dscovr
from . import psp
from . import poes
from . import rbsp
from . import ace
from . import wind
from . import csswe
from . import cluster
from . import geotail
from . import twins
from . import stereo
from . import image
from . import polar
from . import fast
from . import equator_s
from . import solo
from . import secs
from . import kyoto
from . import swarm




from pytplot import get_data, store_data

def tdotp(variable1, variable2, newname=None):
    """
        Routine to calculate the dot product of two tplot variables 
        containing arrays of vectors and storing the results in a 
        tplot variable

    Input
    -------
        variable1: str
            First tplot variable

        variable2: str
            Second tplot variable

    Parameters
    -----------
        newname: str
            Name of the output variable
            
    Returns
    --------
        Name of the tplot variable
    """

    data1 = get_data(variable1, xarray=True)
    data2 = get_data(variable2, xarray=True)

    if data1 is None:
        print('Variable not found: ' + variable1)
        return

    if data2 is None:
        print('Variable not found: ' + variable2)
        return

    if newname is None:
        newname = variable1 + '_dot_' + variable2

    # calculate the dot product
    out = data1.dot(data2, dims='v_dim')

    # save the output
    saved = store_data(newname, data={'x': data1.time.values, 'y': out.values})

    return newname



import numpy as np
from pyspedas import tinterpol
from pyspedas.analysis.lingradest import  lingradest
from pytplot import get_data, store_data, options, join_vec

def mms_lingradest(fields=None, positions=None, suffix=''):
    """

    """

    if fields is None or positions is None:
        print('B-field and spacecraft position keywords required.')
        return

    # interpolate the magnetic field data all onto the same timeline (MMS1):
    # should be in GSE coordinates
    tinterpol(fields[1], fields[0], newname=fields[1] + '_i')
    tinterpol(fields[2], fields[0], newname=fields[2] + '_i')
    tinterpol(fields[3], fields[0], newname=fields[3] + '_i')

    # interpolate the definitive ephemeris onto the magnetic field timeseries
    # should be in GSE coordinates
    tinterpol(positions[0], fields[0], newname=positions[0] + '_i')
    tinterpol(positions[1], fields[0], newname=positions[1] + '_i')
    tinterpol(positions[2], fields[0], newname=positions[2] + '_i')
    tinterpol(positions[3], fields[0], newname=positions[3] + '_i')

    B1 = get_data(fields[0])
    B2 = get_data(fields[1] + '_i')
    B3 = get_data(fields[2] + '_i')
    B4 = get_data(fields[3] + '_i')

    Bx1 = B1.y[:, 0]
    By1 = B1.y[:, 1]
    Bz1 = B1.y[:, 2]

    Bx2 = B2.y[:, 0]
    By2 = B2.y[:, 1]
    Bz2 = B2.y[:, 2]

    Bx3 = B3.y[:, 0]
    By3 = B3.y[:, 1]
    Bz3 = B3.y[:, 2]

    Bx4 = B4.y[:, 0]
    By4 = B4.y[:, 1]
    Bz4 = B4.y[:, 2]

    R1 = get_data(positions[0] + '_i')
    R2 = get_data(positions[1] + '_i')
    R3 = get_data(positions[2] + '_i')
    R4 = get_data(positions[3] + '_i')

    # start the calculation
    output = lingradest(Bx1, Bx2, Bx3, Bx4,
                        By1, By2, By3, By4,
                        Bz1, Bz2, Bz3, Bz4,
                        R1.y, R2.y, R3.y, R4.y)
    # end of the calculations

    # store the results
    store_data('Bt' + suffix, data={'x': B1.times, 'y': output['Bbc']})
    store_data('Bx' + suffix, data={'x': B1.times, 'y': output['Bxbc']})
    store_data('By' + suffix, data={'x': B1.times, 'y': output['Bybc']})
    store_data('Bz' + suffix, data={'x': B1.times, 'y': output['Bzbc']})

    join_vec(['Bt'+suffix, 'Bx'+suffix, 'By'+suffix, 'Bz'+suffix], new_tvar='Bbc' + suffix)

    # B-field gradients
    store_data('gradBx' + suffix, data={'x': B1.times, 'y': output['LGBx']})
    store_data('gradBy' + suffix, data={'x': B1.times, 'y': output['LGBy']})
    store_data('gradBz' + suffix, data={'x': B1.times, 'y': output['LGBz']})

    CB = np.sqrt(output['LCxB']**2 + output['LCyB']**2 + output['LCzB']**2)

    # in nT/1000km
    store_data('absCB' + suffix, data={'x': B1.times, 'y': CB})
    store_data('CxB' + suffix, data={'x': B1.times, 'y': output['LCxB']})
    store_data('CyB' + suffix, data={'x': B1.times, 'y': output['LCyB']})
    store_data('CzB' + suffix, data={'x': B1.times, 'y': output['LCzB']})

    store_data('divB_nT/1000km' + suffix, data={'x': B1.times, 'y': output['LD']})

    join_vec(['absCB'+suffix, 'CxB'+suffix, 'CyB'+suffix, 'CzB'+suffix], new_tvar='curlB_nT/1000km' + suffix)

    # jx in nA/m^2
    store_data('jx' + suffix, data={'x': B1.times, 'y': 0.8 * output['LCxB']})
    # jy in nA/m^2
    store_data('jy' + suffix, data={'x': B1.times, 'y': 0.8 * output['LCyB']})
    # jz in nA/m^2
    store_data('jz' + suffix, data={'x': B1.times, 'y': 0.8 * output['LCzB']})

    join_vec(['jx'+suffix, 'jy'+suffix, 'jz'+suffix], new_tvar='jtotal' + suffix)

    store_data('curvx' + suffix, data={'x': B1.times, 'y': output['curv_x_B']})
    store_data('curvy' + suffix, data={'x': B1.times, 'y': output['curv_y_B']})
    store_data('curvz' + suffix, data={'x': B1.times, 'y': output['curv_z_B']})

    join_vec(['curvx'+suffix,  'curvy'+suffix,  'curvz'+suffix], new_tvar='curvB' + suffix)

    store_data('Rc_1000km' + suffix, data={'x': B1.times, 'y': output['RcurvB']})


import numpy as np


def slice2d_subtract(vectors=None, velocity=None):
    """
    Shift velocities by specified vector
    """

    if vectors is None or velocity is None:
        return

    if vectors.shape[1] != 3 or len(velocity) != 3:
        print('Invalid vector dimensions, cannot subtract velocity')
        return

    if np.sum(~np.isfinite((velocity))):
        print('Invalid bulk velocity data, cannot subtract velocity')
        return

    if not isinstance(velocity, np.ndarray):
        velocity = np.array(velocity)

    v = -velocity

    vectors[:, 0] = vectors[:, 0] + v[0]
    vectors[:, 1] = vectors[:, 1] + v[1]
    vectors[:, 2] = vectors[:, 2] + v[2]

    return vectors




import numpy as np
import pytplot

def data_exists(tvar):
    """
    Checks if a tplot variable exists
    """
    if tvar in pytplot.data_quants.keys():
        data = pytplot.get_data(tvar)
        # multi-dimensional data returns a tuple, NRV variables return an ndarray
        if isinstance(data, tuple) or isinstance(data, np.ndarray) or isinstance(data, str) or isinstance(data, list):
            return True
    return False




from pyspedas.mms.mms_load_data import mms_load_data
from pyspedas.mms.fgm.mms_fgm_remove_flags import mms_fgm_remove_flags
from pyspedas.mms.fgm.mms_fgm_set_metadata import mms_fgm_set_metadata
from pyspedas.mms.fgm.mms_split_fgm_data import mms_split_fgm_data
from pyspedas.mms.print_vars import print_vars
from pyspedas.mms.mms_config import CONFIG
from pyspedas.utilities.data_exists import data_exists

from pytplot import del_data

import re

@print_vars
def mms_load_fgm(trange=['2015-10-16', '2015-10-17'], probe='1', data_rate='srvy',
    level='l2', instrument='fgm', datatype='', varformat=None, varnames=[], suffix='',
    keep_flagged=False, get_support_data=True, time_clip=False, no_update=False,
    available=False, notplot=False, latest_version=False, major_version=False, 
    min_version=None, cdf_version=None, spdf=False, always_prompt=False, no_split_vars=False,
    get_fgm_ephemeris=False):
    """
    This function loads FGM data into tplot variables
    
    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe : str or list of str
            list of probes, valid values for MMS probes are ['1','2','3','4']. 

        data_rate : str or list of str
            instrument data rates for FGM include 'brst' 'fast' 'slow' 'srvy'. The
            default is 'srvy'.

        level : str
            indicates level of data processing. the default if no level is specified is 'l2'

        datatype : str or list of str
            no datatype for FGM instrument (all science data are loaded)

        get_support_data: bool
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a 
            "VAR_TYPE" attribute of "data".

        time_clip: bool
            Data will be clipped to the exact trange specified by the trange keyword.
            
        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        notplot: bool
            If True, then data are returned in a hash table instead of 
            being stored in tplot variables (useful for debugging, and
            access to multi-dimensional data products)

        available: bool
            If True, simply return the available data files (without downloading)
            for the requested paramters

        no_update: bool
            Set this flag to preserve the original data. if not set and newer 
            data is found the existing data will be overwritten

        cdf_version: str
            Specify a specific CDF version # to load (e.g., cdf_version='4.3.0')

        min_version: str
            Specify a minimum CDF version # to load

        latest_version: bool
            Only grab the latest CDF version in the requested time interval

        major_version: bool
            Only open the latest major CDF version (e.g., X in vX.Y.Z) in the requested time interval

        keep_flagged: bool
            If True, don't remove flagged data (flagged data are set to NaNs by
            default, this keyword turns this off)

        always_prompt: bool
            Set this keyword to always prompt for the user's username and password;
            useful if you accidently save an incorrect password, or if your SDC password has changed

        spdf: bool
            If True, download the data from the SPDF instead of the SDC

        get_fgm_ephemeris: bool
            Keep the ephemeris variables in the FGM files
            
    Returns:
        List of tplot variables created.

    """

    if (varformat is not None) and (not keep_flagged) and (not available) and (not notplot):
        varformat_fetch = varformat+'|*_flag_*'
    else:
        varformat_fetch = varformat

    tvars = mms_load_data(trange=trange, notplot=notplot, probe=probe, data_rate=data_rate, level=level, instrument=instrument,
            datatype=datatype, varformat=varformat_fetch, varnames=varnames, suffix=suffix, get_support_data=get_support_data,
            time_clip=time_clip, no_update=no_update, available=available, latest_version=latest_version, major_version=major_version, 
            min_version=min_version, cdf_version=cdf_version, spdf=spdf, always_prompt=always_prompt)
    
    if tvars is None or available or notplot or CONFIG['download_only']:
        return tvars

    if not isinstance(probe, list): probe = [probe]
    if not isinstance(data_rate, list): data_rate = [data_rate]
    if not isinstance(level, list): level = [level]

    # the probes will need to be strings beyond this point
    if isinstance(probe, list):
        probe = [str(p) for p in probe]

    # remove flagged data
    if not keep_flagged:
        mms_fgm_remove_flags(probe, data_rate, level, instrument, suffix=suffix)
        # Delete the flags variable if it was not originally requested
        if varformat is not None:
            regex = re.compile(varformat.replace("*", ".*"))
            tvars_to_delete = [tvar for tvar in tvars if not re.match(regex, tvar)]
            for tvar in tvars_to_delete:
                del_data(tvar)
                tvars.remove(tvar)

    for prb in probe:
        for drate in data_rate:
            for lvl in level:
                if not no_split_vars:
                    out = mms_split_fgm_data(prb, drate, lvl, instrument, suffix=suffix)
                    tvars.extend(out)

                if lvl.lower() != 'ql':
                    # delete the ephemeris variables if not requested
                    if not get_fgm_ephemeris:
                        if data_exists('mms'+prb+'_'+instrument+'_r_gse_'+drate+'_'+lvl+suffix):
                            del_data('mms'+prb+'_'+instrument+'_r_gse_'+drate+'_'+lvl+suffix)
                            tvars.remove('mms'+prb+'_'+instrument+'_r_gse_'+drate+'_'+lvl+suffix)
                        if data_exists('mms'+prb+'_'+instrument+'_r_gsm_'+drate+'_'+lvl+suffix):
                            del_data('mms'+prb+'_'+instrument+'_r_gsm_'+drate+'_'+lvl+suffix)
                            tvars.remove('mms'+prb+'_'+instrument+'_r_gsm_'+drate+'_'+lvl+suffix)
                        if data_exists('mms'+prb+'_pos_gse'+suffix):
                            del_data('mms'+prb+'_pos_gse'+suffix)
                            tvars.remove('mms'+prb+'_pos_gse'+suffix)
                        if data_exists('mms'+prb+'_pos_gsm'+suffix):
                            del_data('mms'+prb+'_pos_gsm'+suffix)
                            tvars.remove('mms'+prb+'_pos_gsm'+suffix)

    mms_fgm_set_metadata(probe, data_rate, level, instrument, suffix=suffix)

    return tvars



import pyspedas
import pytplot
import numpy as np
from pyspedas.themis.spacecraft.fields.fit import cal_fit
from pyspedas.utilities.download import download
from pyspedas.themis.config import CONFIG

# Download tplot files
remote_server = 'https://spedas.org/'
remote_name ='testfiles/cal_fit.tplot'
calfile = download(remote_file=remote_name,
                   remote_path=remote_server,
                   local_path=CONFIG['local_data_dir'],
                   no_download=False)

remote_name ='testfiles/tha_efs_no_cal.tplot'
nocalfile = download(remote_file=remote_name,
                   remote_path=remote_server,
                   local_path=CONFIG['local_data_dir'],
                   no_download=False)

# Load validation variables
filename = calfile[0]
pytplot.tplot_restore(filename)
tha_fit_idl = pytplot.get_data('tha_fit')
tha_fgs_idl = pytplot.get_data('tha_fgs')
tha_fgs_sigma_idl = pytplot.get_data('tha_fgs_sigma')
tha_fit_bfit_idl = pytplot.get_data('tha_fit_bfit')
tha_fit_efit_idl = pytplot.get_data('tha_fit_efit')
tha_efs_idl = pytplot.get_data('tha_efs')
tha_efs_sigma_idl = pytplot.get_data('tha_efs_sigma')
tha_efs_0_idl = pytplot.get_data('tha_efs_0')
tha_efs_dot0_idl = pytplot.get_data('tha_efs_dot0')

pytplot.tplot_names()
pytplot.del_data('*')

filename = nocalfile[0]
pytplot.tplot_restore(filename)
tha_efs_idl_no_cal = pytplot.get_data('tha_efs')

pytplot.tplot_names()
pytplot.del_data('*')

t = ['2008-03-15', '2008-03-16']
pyspedas.themis.fit(trange=t, probe='a', level='l1', get_support_data=True,
                    varnames=['tha_fit', 'tha_fit_code', 'tha_fit_npts'], time_clip=True)
# cal_fit
cal_fit(probe='a')
tha_fit = pytplot.get_data('tha_fit')

tha_fgs = pytplot.get_data('tha_fgs')
tha_fgs_sigma = pytplot.get_data('tha_fgs_sigma')

tha_fit_bfit = pytplot.get_data('tha_fit_bfit')
tha_fit_efit = pytplot.get_data('tha_fit_efit')

tha_efs = pytplot.get_data('tha_efs')
tha_efs_sigma = pytplot.get_data('tha_efs_sigma')

tha_efs_0 = pytplot.get_data('tha_efs_0')
tha_efs_dot0 = pytplot.get_data('tha_efs_dot0')

# no_cal flag
cal_fit(probe='a', no_cal=True)
tha_efs_no_cal = pytplot.get_data('tha_efs')

# Comparison
diff_fgs = np.nanmedian(tha_fgs.y - tha_fgs_idl.y, axis=0, keepdims=True)
diff_sigma = np.nanmedian(tha_fgs_sigma.y - tha_fgs_sigma_idl.y,  axis=0, keepdims=True)
diff_bfit = np.nanmedian(tha_fit_bfit.y - tha_fit_bfit_idl.y,  axis=0, keepdims=True)
diff_efit = np.nanmedian(tha_fit_efit.y - tha_fit_efit_idl.y,  axis=0, keepdims=True)
diff_efs = np.nanmedian(tha_efs.y - tha_efs_idl.y,  axis=0, keepdims=True)
diff_efs_sigma = np.nanmedian(tha_efs_sigma.y - tha_efs_sigma_idl.y,  axis=0, keepdims=True)
diff_efs_no_cal = np.nanmedian(tha_efs_no_cal.y - tha_efs_idl_no_cal.y,  axis=0, keepdims=True)

diff_efs_0 = np.nanmedian(tha_efs_0.y - tha_efs_0_idl.y,  axis=0, keepdims=True)
diff_efs_dot0 = np.nanmedian(tha_efs_dot0.y - tha_efs_dot0_idl.y,  axis=0, keepdims=True)


print('fgs: ', diff_fgs)
print('fgs_sigma:', diff_sigma)

print('bfit:', diff_bfit)
print('efit:', diff_efit)

print('efs:', diff_efs)
print('efs_sigma:', diff_efs_sigma)

print('efs_0:', diff_efs_0)
print('efs_dot0:', diff_efs_dot0)

print('efs_no_cal:', diff_efs_no_cal)



"""
This module contains routines for loading MMS data

"""

from functools import wraps

from .mec_ascii.state import mms_load_state
from .fgm.fgm import mms_load_fgm
from .hpca.hpca import mms_load_hpca
from .fpi.fpi import mms_load_fpi
from .scm.scm import mms_load_scm
from .mec.mec import mms_load_mec
from .feeps.feeps import mms_load_feeps
from .eis.eis import mms_load_eis
from .edi.edi import mms_load_edi
from .edp.edp import mms_load_edp
from .dsp.dsp import mms_load_dsp
from .aspoc.aspoc import mms_load_aspoc
from .fsm.fsm import mms_load_fsm
from .fgm.mms_curl import mms_curl
from .fgm.mms_lingradest import mms_lingradest
from .spd_mms_load_bss import spd_mms_load_bss

'''
    the following wrappers allow users to import the load routines using 
    the syntax: 
    
            >>> from pyspedas.mms import fgm
            >>> fgm_data = fgm(...)

        and/or

            >>> import pyspedas
            >>> fgm_data = pyspedas.mms.fgm(...)
'''

@wraps(spd_mms_load_bss)
def bss(*args, **kwargs):
    return spd_mms_load_bss(*args, **kwargs)

@wraps(mms_load_state)
def state(*args, **kwargs):
    return mms_load_state(*args, **kwargs)

@wraps(mms_load_fgm)
def fgm(*args, **kwargs):
    return mms_load_fgm(*args, **kwargs)

@wraps(mms_load_scm)
def scm(*args, **kwargs):
    return mms_load_scm(*args, **kwargs)

@wraps(mms_load_fsm)
def fsm(*args, **kwargs):
    return mms_load_fsm(*args, **kwargs)

@wraps(mms_load_edp)
def edp(*args, **kwargs):
    return mms_load_edp(*args, **kwargs)

@wraps(mms_load_edi)
def edi(*args, **kwargs):
    return mms_load_edi(*args, **kwargs)

@wraps(mms_load_fpi)
def fpi(*args, **kwargs):
    return mms_load_fpi(*args, **kwargs)

@wraps(mms_load_hpca)
def hpca(*args, **kwargs):
    return mms_load_hpca(*args, **kwargs)

@wraps(mms_load_eis)
def eis(*args, **kwargs):
    return mms_load_eis(*args, **kwargs)

@wraps(mms_load_feeps)
def feeps(*args, **kwargs):
    return mms_load_feeps(*args, **kwargs)

@wraps(mms_load_aspoc)
def aspoc(*args, **kwargs):
    return mms_load_aspoc(*args, **kwargs)

@wraps(mms_load_mec)
def mec(*args, **kwargs):
    return mms_load_mec(*args, **kwargs)

@wraps(mms_load_dsp)
def dsp(*args, **kwargs):
    return mms_load_dsp(*args, **kwargs)

@wraps(mms_curl)
def curlometer(*args, **kwargs):
    return mms_curl(*args, **kwargs)

@wraps(mms_lingradest)
def lingradest(*args, **kwargs):
    return mms_lingradest(*args, **kwargs)




import logging
import warnings
import numpy as np
from pytplot import get_data, store_data, options
from pyspedas import tnames

# use nanmean from bottleneck if it's installed, otherwise use the numpy one
# bottleneck nanmean is ~2.5x faster
try:
    import bottleneck as bn
    nanmean = bn.nanmean
except ImportError:
    nanmean = np.nanmean

logging.captureWarnings(True)
logging.basicConfig(format='%(asctime)s: %(message)s', datefmt='%d-%b-%y %H:%M:%S', level=logging.INFO)

def mms_eis_spin_avg(probe='1', species='proton', data_units='flux', datatype='extof', data_rate='srvy', level='l2', suffix=''):
    """
    This function will spin-average the EIS spectrograms, and is automatically called from mms_load_eis
    
    Parameters
    ----------
        probe: str
            probe #, e.g., '4' for MMS4

        species: str
            species for calculation (default: 'proton')
            
        data_units: str
            'flux' or 'cps' (default: 'flux')

        datatype: str
            'extof' or 'phxtof' (default: 'extof')

        data_rate: str
            instrument data rate, e.g., 'srvy' or 'brst' (default: 'srvy')

        level: str
            data level ['l1a','l1b','l2pre','l2' (default)]

        suffix: str
            suffix of the loaded data

    Returns:
        List of tplot variables created.
    """
    prefix = 'mms' + probe + '_epd_eis_' + data_rate + '_' + level + '_'

    if data_units == 'flux':
        units_label = '1/(cm^2-sr-s-keV)'
    elif data_units == 'cps':
        units_label = '1/s'
    elif data_units == 'counts':
        units_label = 'counts'

    spin_data = get_data(prefix + datatype + '_spin' + suffix)
    if spin_data is None:
        logging.error('Error, problem finding EIS spin variable to calculate spin-averages')
        return

    spin_times, spin_nums = spin_data

    if spin_nums is not None:
        spin_starts = [spin_start for spin_start in np.where(spin_nums[1:] > spin_nums[:-1])[0]]

        telescopes = tnames(prefix + datatype + '_' + species + '_*' + data_units + '_t?' + suffix)

        if len(telescopes) != 6:
            logging.error('Problem calculating the spin-average for species: ' + species + ' (' + datatype + ')')
            return None

        out_vars = []

        for scope in range(0, 6):
            this_scope = telescopes[scope]
            scope_data = get_data(this_scope)
            
            if len(scope_data) <= 2:
                logging.error("Error, couldn't find energy table for the variable: " + this_scope)
                continue

            flux_times, flux_data, energies = scope_data

            spin_avg_flux = np.zeros([len(spin_starts), len(energies)])

            current_start = 0

            for spin_idx in range(0, len(spin_starts)):
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", category=RuntimeWarning)
                    spin_avg_flux[spin_idx, :] = nanmean(flux_data[current_start:spin_starts[spin_idx]+1, :], axis=0)
                current_start = spin_starts[spin_idx] + 1

            store_data(this_scope + '_spin', data={'x': flux_times[spin_starts], 'y': spin_avg_flux, 'v': energies})
            options(this_scope + '_spin', 'ztitle', units_label)
            options(this_scope + '_spin', 'ytitle', 'MMS' + probe + ' ' + datatype + ' ' + species + ' (spin)')
            options(this_scope + '_spin', 'ysubtitle', 'Energy [keV]')
            options(this_scope + '_spin', 'spec', True)
            options(this_scope + '_spin', 'ylog', True)
            options(this_scope + '_spin', 'zlog', True)
            options(this_scope + '_spin', 'Colormap', 'spedas')
            out_vars.append(this_scope + '_spin')
        return out_vars
    else:
        logging.error('Error, problem finding EIS spin variable to calculate spin-averages')
        return None


from pytplot import get_data, store_data
from pyspedas import cotrans_set_coord, tinterpol

try:
    import spacepy.coordinates as coord
except ImportError:
    print('SpacePy must be installed to use this module.')
    print('Please install it using: pip install spacepy')


def mms_cotrans_qrotate(in_name, q_name, out_name, out_coord, inverse=False):
    """
    Performs a quaternion rotation on a tplot variable.

    Note: this routine will interpolate the input data to the quaternion
    timestamps if the data aren't the same length as the quaternions

    Parameters
    ----------
        in_name: str
            Tplot variable to be transformed

        q_name: str
            Tplot variable containing the quaternions

        out_name: str
            Tplot variable to be created

        inverse: bool:
            Flag to use the quaternionConjugate on the quaternion data prior to rotating
    """
    data = get_data(in_name)
    metadata = get_data(in_name, metadata=True)

    q_data = get_data(q_name)

    if len(data.times) != len(q_data.times):
        print('Interpolating the data to the MEC quaternion time stamps.')
        tinterpol(in_name, q_name)
        data = get_data(in_name + '-itrp')

    if inverse:
        quaternion = coord.quaternionConjugate(q_data.y)
    else:
        quaternion = q_data.y

    out_data = coord.quaternionRotateVector(quaternion, data.y)

    saved = store_data(out_name, data={'x': data.times, 'y': out_data}, attr_dict=metadata)
    if saved:
        cotrans_set_coord(out_name, out_coord)





import csv
import logging
import requests
import numpy as np
from pytplot import store_data, options
from pyspedas import time_double, time_string

logging.captureWarnings(True)
logging.basicConfig(format='%(asctime)s: %(message)s', datefmt='%d-%b-%y %H:%M:%S', level=logging.INFO)

def get_mms_srois(start_time=None, end_time=None, sc_id=None):
    if start_time is None:
        logging.error('Error, start time not specified')
        return

    if end_time is None:
        logging.error('Error, end time not specified')
        return

    if sc_id is None:
        logging.error('Error, sc_id not specified')
        return

    # public unauthenticated path
    path = 'mms/sdc/public/service/latis/mms_events_view.csv'

    query = '?start_time_utc,end_time_utc,sc_id,start_orbit&event_type=SROI'

    # convert to standard ISO format as accepted by LaTiS
    query += '&start_time_utc>=' + 'T'.join(start_time.split(' '))
    query += '&start_time_utc<' + 'T'.join(end_time.split(' '))
    query += '&sc_id=' + sc_id.lower()

    qreq = requests.get('https://lasp.colorado.edu/' + path + query)

    if qreq.status_code != 200:
        logging.error('Error downloading SRoI segments')
        return

    sroi_data = csv.reader(qreq.content.decode('utf-8').splitlines())

    count = 0
    unix_start = []
    unix_end = []

    for line in sroi_data:
        if count == 0:
            # ignore the first line
            count += 1
            continue

        if len(line) != 4:
            continue

        unix_start.append(time_double(line[0]))
        unix_end.append(time_double(line[1]))

        count += 1

    return (np.array(unix_start), np.array(unix_end))


def mms_load_sroi_segments(trange=None, probe=1, suffix=''):
    """
    This function loads the Science Region of Interest (SRoI) segment intervals
    
    Parameters:
        trange: list of str
            time range of interest [starttime, endtime] with the format 
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day 
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe: str or int
            Spacecraft probe # (default: 1)

        suffix: str
            Suffix to append to the end of the tplot variables
            
    Returns:
        Tuple containing (start_times, end_times)

    """

    if not isinstance(probe, str):
        probe = str(probe)

    if trange is None:
        logging.error('Error; no trange specified.')
        return None

    tr = time_double(trange)

    unix_start, unix_end = get_mms_srois(
        start_time=time_string(tr[0]-2*86400.0, fmt='%Y-%m-%d %H:%M:%S'), 
        end_time=time_string(tr[1]+2*86400.0, fmt='%Y-%m-%d %H:%M:%S'),
        sc_id='mms'+str(probe))

    if len(unix_start) == 0:
        return ([], [])

    times_in_range = (unix_start >= tr[0]-2*86400.0) & (unix_start <= tr[1]+2*86400.0)

    unix_start = unix_start[times_in_range]
    unix_end = unix_end[times_in_range]

    start_out = []
    end_out = []
    bar_x = []
    bar_y = []

    for start_time, end_time in zip(unix_start, unix_end):
        if end_time >= tr[0] and start_time <= tr[1]:
            bar_x.extend([start_time, start_time, end_time, end_time])
            bar_y.extend([np.nan, 0., 0., np.nan])
            start_out.append(start_time)
            end_out.append(end_time)

    vars_created = store_data('mms' + probe + '_bss_sroi'+suffix, data={'x': bar_x, 'y': bar_y})

    if not vars_created:
        logging.error('Error creating SRoI segment intervals tplot variable')
        return None

    options('mms' + probe + '_bss_sroi'+suffix, 'panel_size', 0.09)
    options('mms' + probe + '_bss_sroi'+suffix, 'thick', 2)
    options('mms' + probe + '_bss_sroi'+suffix, 'Color', 'green')
    options('mms' + probe + '_bss_sroi'+suffix, 'border', False)
    options('mms' + probe + '_bss_sroi'+suffix, 'yrange', [-0.001,0.001])
    options('mms' + probe + '_bss_sroi'+suffix, 'legend_names', ['Fast'])
    options('mms' + probe + '_bss_sroi'+suffix, 'ytitle', '')

    return (start_out, end_out)




import numpy as np
import pandas as pd
import zipfile
from tempfile import mkdtemp
from pyspedas import time_double
from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pytplot import store_data

def get_w(trange=None, create_tvar=False, newname=None):
    """
    This routine downloads the 6 Tsygeneko (TS05) model 
    driving variables W1, W2, W3, W4, W5, W6; from:

    http://geo.phys.spbu.ru/~tsyganenko/TS05_data_and_stuff
    """

    if trange is None:
        print('trange keyword must be specified.')
        return

    years = dailynames(trange=trange, file_format='%Y')
    tmpdir = mkdtemp()

    if newname is None:
        newname = 'Tsy_W_vars_' + '-'.join(years)

    ut_out = np.empty(0)
    w1_out = np.empty(0)
    w2_out = np.empty(0)
    w3_out = np.empty(0)
    w4_out = np.empty(0)
    w5_out = np.empty(0)
    w6_out = np.empty(0)

    for year in years:
        file = download(remote_path='http://geo.phys.spbu.ru/~tsyganenko/TS05_data_and_stuff/',
                        remote_file=year+'_OMNI_5m_with_TS05_variables.???',
                        local_path=tmpdir)

        if file[0][-3:] == 'zip':
            with zipfile.ZipFile(file[0], 'r') as zip_ref:
                zip_ref.extractall(tmpdir)

        rows = pd.read_csv(tmpdir + '/' + year + '_OMNI_5m_with_TS05_variables.dat', delim_whitespace=True, header=None)

        # extract the W variables
        w1 = rows.to_numpy()[:, -6]
        w2 = rows.to_numpy()[:, -5]
        w3 = rows.to_numpy()[:, -4]
        w4 = rows.to_numpy()[:, -3]
        w5 = rows.to_numpy()[:, -2]
        w6 = rows.to_numpy()[:, -1]

        # extract the times
        years = rows.to_numpy()[:, 0]
        doys = rows.to_numpy()[:, 1]
        hours = rows.to_numpy()[:, 2]
        minutes = rows.to_numpy()[:, 3]
        time_strings = [str(int(year)) + '-' + str(int(doy)).zfill(3) + ' ' + str(int(hour)).zfill(2) + ':' + str(int(minute)).zfill(2) for year, doy, hour, minute in zip(years, doys, hours, minutes)]
        unix_times = np.array(time_double(time_strings))

        ut_out = np.append(ut_out, unix_times)
        w1_out = np.append(w1_out, w1)
        w2_out = np.append(w2_out, w2)
        w3_out = np.append(w3_out, w3)
        w4_out = np.append(w4_out, w4)
        w5_out = np.append(w5_out, w5)
        w6_out = np.append(w6_out, w6)

    in_range = np.argwhere((ut_out >= time_double(trange[0])) & (ut_out < time_double(trange[1]))).squeeze()

    if len(in_range) == 0:
        print('No data found in the trange.')
        return

    if create_tvar:
        out = np.array((w1_out[in_range], w2_out[in_range], w3_out[in_range], w4_out[in_range], w5_out[in_range], w6_out[in_range]))
        store_data(newname, data={'x': ut_out[in_range], 'y': out.T})
        return newname

    return {'times': ut_out[in_range], 
            'w1': w1_out[in_range], 
            'w2': w2_out[in_range], 
            'w3': w3_out[in_range], 
            'w4': w4_out[in_range], 
            'w5': w5_out[in_range], 
            'w6': w6_out[in_range]}



import numpy as np
from pytplot import get_data, store_data

def tcrossp(v1, v2, newname=None, return_data=False):
    """
    Calculates the cross product of two tplot varibles

    Input
    -------
        v1: str
            First tplot variable

        v2: str
            Second tplot variable

    Parameters
    -----------
        newname: str
            Name of the output variable

        return_data: bool
            Returns the data as an ndarray instead of creating a tplot variable

    Returns
    --------
        Name of the tplot variable
    """

    v1_data = None
    v2_data = None

    if not isinstance(v1, np.ndarray) and isinstance(v1, str):
        v1_data = get_data(v1)
        v1_name = v1

        if v1_data is not None:
            data1 = v1_data[1]
    else:
        v1_name = 'var1'
        data1 = v1


    if not isinstance(v2, np.ndarray) and isinstance(v2, str):
        v2_data = get_data(v2)
        v2_name = v2

        if v2_data is not None:
            data2 = v2_data[1]
    else:
        v2_name = 'var2'
        data2 = v2

    if newname is None:
        newname = v1_name + '_cross_' + v2_name

    cp = np.cross(data1, data2)

    if return_data:
        return cp
    else:
        out = cp
        if v2_data is None:
            if len(cp.shape) == 1:
                out = np.atleast_2d(cp)
            times = np.zeros(out.shape[0])
        else:
            times = v2_data[0]
        store_data(newname, data={'x': times, 'y': out})
        return newname




'''

This script loads and prints FGM data at several data points.

This is meant to be called from the IDL test suite for comparison with the data loaded via IDL SPEDAS

'''


import pyspedas
from pytplot import get_data

thm_vars = pyspedas.themis.fgm(probe='c', level='l2')

data = get_data('thc_fgs_dsl')

print(data[0][0:10].round(6).tolist())

print(data[1][1000].tolist())

print(data[1][5000].tolist())

print(data[1][10000].tolist())

print(data[1][20000].tolist())



import numpy as np
import scipy.signal as signal


def slice2d_smooth(the_slice, width):
    """
    Smooths the output data by applying a gaussian blur.
    """

    smooth = np.round(width)

    if smooth >= 2:
        # ensure kernel with odd # of elements in [3:slice resolution]
        if smooth > 3:
            n = smooth

        if the_slice['data'].shape[0] < 3:
            n = the_slice['data'].shape[0]

        if n % 2 == 0:
            n = n - 1

        # Increase the STDV with the smoothing width to allow convolve2d to
        # effectively smooth over a greater area. Simply extending the
        # width with constant STDV adds negligable terms to the kernel.
        # Allowing the STDV to grow too quickly results in flat distributions.
        c = np.log(n) - 0.25  # There's no simple solution for c = f(n) for small n
                              # so this was partly arrived at empirically. It
                              # should allow the width of the gaussian to increase
                              # with the smoothing window without becoming flat
                              # (too quick) or adding negligible terms (to slow).

        # set up the kernel
        s = np.arange(0, n) - np.floor(n/2.0)
        ones = np.ones(n)
        sx = np.outer(s, ones)
        sy = np.outer(s, ones).T

        # 2D normalized gaussian kernel
        kernel = np.exp((-0.5)*((sx/c)**2 + (sy/c)**2))
        kernel = kernel / np.nansum(kernel)

        the_slice['data'] = signal.convolve2d(the_slice['data'], kernel, mode='same')
    else:
        print('Smoothing not applied. Smoothing value must be >= 2')

    return the_slice



from pyspedas import tnames
from pytplot import get_data, store_data, options

def mms_load_fpi_calc_pad(probe='1', level='sitl', datatype='', data_rate='', suffix='', autoscale=True):
    """
    Calculates the omni-directional pitch angle distribution (summed and averaged)
    from the individual tplot variables
    
    Parameters
    ----------
        probe: str 
            probe, valid values for MMS probes are ['1','2','3','4']. 

        level: str
            indicates level of data processing. the default if no level is specified is 'sitl'

        datatype: str
            Valid datatypes for FPI are:
              Quicklook: ['des', 'dis'] 
              SITL: '' (none; loads both electron and ion data from single CDF)
              L1b/L2: ['des-dist', 'dis-dist', 'dis-moms', 'des-moms']

        data_rate: str
            instrument data rates for FPI include 'brst' and 'fast'. The
            default is 'fast'.

        suffix: str
            The tplot variable names will be given this suffix.  By default, 
            no suffix is added.

        autoscale: bool
            If set, use the default zrange; otherwise, use the min and max of the data for the zrange

    Returns:
        List of tplot variables created.

    """
    out_vars = []

    if isinstance(datatype, str):
        if datatype == '*' or datatype == '':
            if level.lower() == 'ql':
                datatype = ['des', 'dis']
            else:
                datatype = ['des-dist', 'dis-dist']

    if isinstance(datatype, str):
        datatype = [datatype]

    for dtype in datatype:
        species = dtype[1]

        if level.lower() == 'sitl':
            spec_str_format = 'PitchAngDist'
            obs_str_format = '_fpi_' + species
        else:
            spec_str_format = 'pitchAngDist'
            obs_str_format = '_d' + species + 's_'

        obsstr = 'mms' + str(probe) + obs_str_format

        if level.lower() == 'l2':
            spec_str_format = 'pitchangdist'
            pad_vars = [obsstr+spec_str_format+'_'+erange+'en_'+data_rate+suffix for erange in ['low', 'mid', 'high']]
        else:
            pad_vars = [obsstr+spec_str_format+'_'+erange+'En'+suffix for erange in ['low', 'mid', 'high']]

        pad_avg_name = obsstr+'PitchAngDist_avg'+suffix

        low_en = get_data(pad_vars[0])
        mid_en = get_data(pad_vars[1])
        high_en = get_data(pad_vars[2])

        if low_en is None or mid_en is None or high_en is None:
            v3_low_pad = tnames(pad_vars[0].lower()+'_'+data_rate)
            v3_mid_pad = tnames(pad_vars[1].lower()+'_'+data_rate)
            v3_high_pad = tnames(pad_vars[2].lower()+'_'+data_rate)
            if v3_low_pad == [] or v3_mid_pad == [] or v3_high_pad == []:
                continue

            low_en = get_data(v3_low_pad[0])
            mid_en = get_data(v3_mid_pad[0])
            high_en = get_data(v3_high_pad[0])
            pad_avg_name = pad_avg_name.lower()

        e_pad_sum = low_en.y+mid_en.y+high_en.y
        e_pad_avg = e_pad_sum/3.0

        if level == 'l2':
            pad_avg_name = pad_avg_name.lower()

        if species == 'e':
            species_str = 'electron'
        elif species == 'i':
            species_str = 'ion'

        if level == 'ql':
            store_data(obsstr+'PitchAngDist_sum'+suffix, data={'x': low_en.times, 'y': e_pad_sum, 'v': low_en.v})
            options(obsstr+'PitchAngDist_sum'+suffix, 'ytitle', 'MMS'+str(probe)+' \\ '+species_str+' \\ PAD \\ SUM')
            options(obsstr+'PitchAngDist_sum'+suffix, 'yrange', [0, 180])
            options(obsstr+'PitchAngDist_sum'+suffix, 'zlog', True)
            options(obsstr+'PitchAngDist_sum'+suffix, 'spec', True)
            options(obsstr+'PitchAngDist_sum'+suffix, 'Colormap', 'spedas')
            out_vars.append(obsstr+'PitchAngDist_sum'+suffix)

        store_data(pad_avg_name, data={'x': low_en.times, 'y': e_pad_avg, 'v': low_en.v})
        options(pad_avg_name, 'ztitle', 'eV/(cm!U2!N s sr eV)')
        options(pad_avg_name, 'ytitle', 'MMS'+str(probe)+' \\ '+species_str+' \\ PAD \\ AVG')
        options(pad_avg_name, 'yrange', [0, 180])
        options(pad_avg_name, 'zlog', True)
        options(pad_avg_name, 'spec', True)
        options(pad_avg_name, 'Colormap', 'spedas')
        out_vars.append(pad_avg_name)

        return out_vars


import logging
from time import time

logging.captureWarnings(True)
logging.basicConfig(format='%(asctime)s: %(message)s', datefmt='%d-%b-%y %H:%M:%S', level=logging.INFO)

def spd_pgs_progress_update(last_update_time=None, current_sample=None, total_samples=None, type_string=None):
    """
    Helper routine prints status message indicating completion percent

    Parameters:
        last_update_time: float
            Last time the status message was printed (unix time)
    
        current_sample: int
            Current sample #

        total_samples: int
            Total number of samples

        type_string: str
            Usually the variable name (defaults to 'Data')

    Returns:
        float containing the last update time

    """

    if current_sample is None or total_samples is None:
        logging.error('Error, both current sample and total # of samples must be specified.')
        return

    if last_update_time is None:
        return time()

    if type_string is None:
        type_string = 'Data'

    if time()-last_update_time > 10:
        logging.info(type_string + ' is ' + str(round(100.0*current_sample/total_samples)) + '% done.')
        last_update_time = time()

    return last_update_time



def spd_units_string(units, units_only=False):
    """
    Return string describing particle data units for labels

    Input:
        units: str
            String describing units

    Returns:
        String containing full units string
    """
    out = ['', 'Unknown', '']
    units = units.lower()

    if units == 'counts':
        out = ['', 'Counts', '']
    elif units == 'rate':
        out = ['Rate (', '#/sec', ')']
    elif units == 'eflux':
        out = ['Energy Flux (', 'eV / sec / $cm^2$ / ster / eV', ')']
    elif units == 'flux':
        out = ['Flux (', '# / sec / $cm^2$ / ster / eV', ')']
    elif units == 'df':
        out = ['f (', '$s^3$ / $cm^3$ / $km^3$', ')']
    elif units == 'df_cm':
        out = ['f (', '$s^3$ / $cm^6$', ')']
    elif units == 'df_km':
        out = ['f (', '$s^3$ / $km^6$', ')']
    elif units == 'e2flux':
        out = ['$Energy^2$ Flux (', '$eV^2$ / sec / $cm^2$ / ster /eV', ')']
    elif units == 'e3flux':
        out = ['$Energy^3$ Flux (', '$eV^3$ / sec / $cm^2$ / ster /eV', ')']

    if units_only:
        return out[1]

    return ''.join(out)



import numpy as np

def mms_pgs_split_hpca(data_in):
    """
    Split hpca elevation bins so that dphi == dtheta.
    This should allow the regrid step for FAC spectra to be skipped in mms_part_products.
    """
    clean_data = data_in.copy()
    clean_data['data'] = np.concatenate((clean_data['data'], clean_data['data']), axis=1)
    clean_data['bins'] = np.concatenate((clean_data['bins'], clean_data['bins']), axis=1)
    clean_data['energy'] = np.concatenate((clean_data['energy'], clean_data['energy']), axis=1)
    clean_data['denergy'] = np.concatenate((clean_data['denergy'], clean_data['denergy']), axis=1)
    clean_data['phi'] = np.concatenate((clean_data['phi'], clean_data['phi']), axis=1)
    clean_data['dphi'] = np.concatenate((clean_data['dphi'], clean_data['dphi']), axis=1)
    clean_data['theta'] = np.concatenate((clean_data['theta']+0.25*clean_data['dtheta'], clean_data['theta']-0.25*clean_data['dtheta']), axis=1)
    clean_data['dtheta'] = np.concatenate((clean_data['dtheta']/2.0, clean_data['dtheta']/2.0), axis=1)
    
    return clean_data



import requests
from pyspedas import time_double
from pyspedas import time_clip as tclip
from pyspedas.utilities.dailynames import dailynames
from pytplot import store_data, options

def dst(trange=None, time_clip=True, remote_data_dir='http://wdc.kugi.kyoto-u.ac.jp/', suffix=''):
    """
    Loads Dst data from the Kyoto servers. 

    Parameters
    -----------
        trange: list of str
            Time range to load

        time_clip: bool
            If set, time the data to the requested trange

        remote_data_dir: str
            Remote data server at Kyoto

        suffix: str
            Suffix to append to the output variable's name

    Acknowledgment
    ----------------
        The DST data are provided by the World Data Center for Geomagnetism, Kyoto, and
        are not for redistribution (http://wdc.kugi.kyoto-u.ac.jp/). Furthermore, we thank
        the geomagnetic observatories (Kakioka [JMA], Honolulu and San Juan [USGS], Hermanus
        [RSA], Alibag [IIG]), NiCT, INTERMAGNET, and many others for their cooperation to
        make the Dst index available.
    
    Returns
    --------
        Name of the tplot variable created.
    """

    if trange is None:
        print('trange keyword required to download data.')
        return

    file_names = dailynames(file_format='%Y%m/index.html', trange=trange)

    if file_names is None:
        print('No files found for this trange.')
        return
        
    times = []
    data = []
    datatype = ''

    # Final files
    for filename in file_names:
        html_text = requests.get(remote_data_dir + 'dst_final/' + filename).text
        file_times, file_data = parse_html(html_text, year=filename[:4], month=filename[4:6])
        times.extend(file_times)
        data.extend(file_data)
        if len(file_times) != 0:
            datatype = 'Final'

    # Provisional files
    for filename in file_names:
        html_text = requests.get(remote_data_dir + 'dst_provisional/' + filename).text
        file_times, file_data = parse_html(html_text, year=filename[:4], month=filename[4:6])
        times.extend(file_times)
        data.extend(file_data)
        if len(file_times) != 0:
            datatype = 'Provisional'

    # Real Time files
    for filename in file_names:
        html_text = requests.get(remote_data_dir + 'dst_realtime/' + filename).text
        file_times, file_data = parse_html(html_text, year=filename[:4], month=filename[4:6])
        times.extend(file_times)
        data.extend(file_data)
        if len(file_times) != 0:
            datatype = 'Real Time'

    if len(times) == 0:
        print('No data found.')
        return

    store_data('kyoto_dst'+suffix, data={'x': times, 'y': data})

    if time_clip:
        tclip('kyoto_dst'+suffix, trange[0], trange[1], suffix='')

    options('kyoto_dst'+suffix, 'ytitle', 'Dst (' + datatype + ')')

    print('**************************************************************************************')
    print('The DST data are provided by the World Data Center for Geomagnetism, Kyoto, and')
    print(' are not for redistribution (http://wdc.kugi.kyoto-u.ac.jp/). Furthermore, we thank')
    print(' the geomagnetic observatories (Kakioka [JMA], Honolulu and San Juan [USGS], Hermanus')
    print(' [RSA], Alibag [IIG]), NiCT, INTERMAGNET, and many others for their cooperation to')
    print(' make the Dst index available.')
    print('**************************************************************************************')

    return 'kyoto_dst'+suffix

def parse_html(html_text, year=None, month=None):
    times = []
    data = []
    # remove all of the HTML before the table
    html_data = html_text[html_text.find('Hourly Equatorial Dst Values'):]
    # remove all of the HTML after the table
    html_data = html_data[:html_data.find('<!-- vvvvv S yyyymm_part3.html vvvvv -->')]
    html_lines = html_data.split('\n')
    data_strs = html_lines[5:]
    # loop over days
    for day_str in data_strs:
        # the first element of hourly_data is the day, the rest are the hourly Dst values
        hourly_data = day_str.split()
        if len(hourly_data[1:]) != 24:
            continue
        for idx, dst_value in enumerate(hourly_data[1:]):
            times.append(time_double(year + '-' + month + '-' + hourly_data[0] + '/' + str(idx) + ':30'))
            data.append(float(dst_value))

    return (times, data)
